"""
Module for reading and writing CPHD files. Support reading CPHD version 0.3 and 1
and writing version 1.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import os
from typing import Union, List, Tuple, Dict, BinaryIO, Optional, Sequence
from collections import OrderedDict
import numbers

import numpy

from sarpy.io.general.utils import is_file_like, is_real_file
from sarpy.io.general.base import BaseReader, BaseWriter, SarpyIOError
from sarpy.io.general.data_segment import DataSegment, NumpyArraySegment, \
    NumpyMemmapSegment
from sarpy.io.general.format_function import ComplexFormatFunction
from sarpy.io.general.slice_parsing import verify_subscript, verify_slice

from sarpy.io.phase_history.base import CPHDTypeReader
from sarpy.io.phase_history.cphd1_elements.CPHD import CPHDType as CPHDType1, \
    CPHDHeader as CPHDHeader1, CPHD_SECTION_TERMINATOR
from sarpy.io.phase_history.cphd0_3_elements.CPHD import CPHDType as CPHDType0_3, \
    CPHDHeader as CPHDHeader0_3
from sarpy.io.phase_history.cphd_schema import get_namespace, get_default_tuple

logger = logging.getLogger(__name__)

_unhandled_version_text = 'Got unhandled CPHD version number `{}`'
_missing_channel_identifier_text = 'Cannot find CPHD channel for identifier `{}`'
_index_range_text = 'index must be in the range `[0, {})`'


class AmpScalingFunction(ComplexFormatFunction):
    __slots__ = (
        '_amplitude_scaling', )
    _allowed_ordering = ('IQ', )

    def __init__(
            self,
            raw_dtype: Union[str, numpy.dtype],
            raw_shape: Optional[Tuple[int, ...]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Tuple[int, ...]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            band_dimension: int = -1,
            amplitude_scaling: Optional[numpy.ndarray] = None):
        """

        Parameters
        ----------
        raw_dtype : str|numpy.dtype
            The raw datatype. Valid options dependent on the value of order.
        raw_shape : None|Tuple[int, ...]
        formatted_shape : None|Tuple[int, ...]
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]
        band_dimension : int
            Which band is the complex dimension, **after** the transpose operation.
        amplitude_scaling : None|numpy.ndarray
            This is here to support the presence of a scaling in CPHD or CRSD usage.
            This requires that `band_dimension` is the final dimension and neither
            `reverse_axes` nor `transpose_axes` is populated.
        """

        ComplexFormatFunction.__init__(
            self, raw_dtype, 'IQ', raw_shape=raw_shape, formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes, band_dimension=band_dimension)
        self._amplitude_scaling = None
        self.set_amplitude_scaling(amplitude_scaling)

    @property
    def amplitude_scaling(self) -> Optional[numpy.ndarray]:
        """
        The scaling multiplier array, for CPHD/CRSD usage.

        Returns
        -------
        Optional[numpy.ndarray]
        """

        return self._amplitude_scaling

    def set_amplitude_scaling(
            self,
            array: Optional[numpy.ndarray]) -> None:
        """
        Set the amplitude scaling array.

        Parameters
        ----------
        array : None|numpy.ndarray

        Returns
        -------
        None
        """

        if array is None:
            self._amplitude_scaling = None
            return

        if self.order != 'IQ':
            raise ValueError(
                'CPHD requires data in IQ order.')

        if not isinstance(array, numpy.ndarray):
            raise ValueError('requires a numpy.ndarray, got {}'.format(type(array)))
        if array.ndim != 1:
            raise ValueError('requires a one dimensional array')
        if array.dtype.name not in ['float32', 'float64']:
            raise ValueError('requires a numpy.ndarray of float32 or 64 dtype, got {}'.format(array.dtype))
        if array.dtype.name != 'float32':
            array = numpy.asarray(array, dtype=numpy.float32)

        self._amplitude_scaling = array
        self._validate_amplitude_scaling()

    def _validate_amplitude_scaling(self) -> None:
        if self._amplitude_scaling is None or self._raw_shape is None:
            return

        if self.band_dimension not in [-1, self.raw_ndim-1]:
            raise ValueError('Use of scaling multiplier requires band is the final dimension')
        if self.transpose_axes is not None or self.reverse_axes is not None:
            raise ValueError('Use of scaling multiplier requires null reverse_axes and transpose_axes')
        if self._amplitude_scaling.size != self.raw_shape[0]:
            raise ValueError(
                'Use of scaling multiplier requires the array length\n\t'
                'and the first dimension of raw_shape match.')

    def _forward_functional_step(
            self,
            data: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        out = ComplexFormatFunction._forward_functional_step(self, data, subscript)

        # NB: subscript is in raw coordinates, but we have verified that
        #   the first dimension is unchanged
        if self._amplitude_scaling is not None:
            out = self._amplitude_scaling[subscript[0]][:, numpy.newaxis] * out

        return out

    def _reverse_functional_step(
            self,
            data: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        # NB: subscript is in formatted coordinates, but we have verified that
        #   transpose_axes is None and band_dimension is the final dimension
        if self._amplitude_scaling is not None:
            data = (1./self._amplitude_scaling[subscript[0]])[:, numpy.newaxis] * data
        if issubclass(self._raw_dtype.type, numbers.Integral):
            data = numpy.rint(data)

        return ComplexFormatFunction._reverse_functional_step(self, data, subscript)

    def validate_shapes(self) -> None:
        ComplexFormatFunction.validate_shapes(self)
        self._validate_amplitude_scaling()


class CPHDDetails(object):
    """
    The basic CPHD element parser.
    """

    __slots__ = (
        '_file_name', '_file_object', '_closed', '_close_after', '_cphd_version', '_cphd_header', '_cphd_meta')

    def __init__(self, file_object: str):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
            The path to or file like object referencing the CPHD file.
        """

        self._closed = False
        self._close_after = None
        self._cphd_version = None
        self._cphd_header = None
        self._cphd_meta = None
        self._file_object = None  # type: Optional[BinaryIO]

        if isinstance(file_object, str):
            if not os.path.exists(file_object) or not os.path.isfile(file_object):
                raise SarpyIOError('path {} does not exist or is not a file'.format(file_object))
            self._file_name = file_object
            self._file_object = open(file_object, 'rb')
            self._close_after = True
        elif is_file_like(file_object):
            self._file_object = file_object
            if hasattr(file_object, 'name') and isinstance(file_object.name, str):
                self._file_name = file_object.name
            else:
                self._file_name = '<file like object>'
            self._close_after = False
        else:
            raise TypeError('Got unsupported input type {}'.format(type(file_object)))

        self._file_object.seek(0, os.SEEK_SET)
        head_bytes = self._file_object.read(10)
        if not isinstance(head_bytes, bytes):
            raise ValueError('Input file like object not open in bytes mode.')
        if not head_bytes.startswith(b'CPHD'):
            raise SarpyIOError('File {} does not appear to be a CPHD file.'.format(self.file_name))

        self._extract_version()
        self._extract_header()
        self._extract_cphd()

    @property
    def file_name(self) -> str:
        """
        str: The CPHD filename.
        """

        return self._file_name

    @property
    def file_object(self) -> BinaryIO:
        """
        BinaryIO: The binary file object
        """
        # noinspection PyTypeChecker
        return self._file_object

    @property
    def cphd_version(self) -> str:
        """
        str: The CPHD version.
        """

        return self._cphd_version

    @property
    def cphd_meta(self) -> Union[CPHDType1, CPHDType0_3]:
        """
        CPHDType1|CPHDType0_3: The CPHD metadata object, which is version dependent.
        """

        return self._cphd_meta

    @property
    def cphd_header(self) -> Union[CPHDHeader1, CPHDHeader0_3]:
        """
        CPHDHeader1|CPHDHeader0_3: The CPHD header object, which is version dependent.
        """

        return self._cphd_header

    def _extract_version(self) -> None:
        """
        Extract the version number from the file. This will advance the file
        object to the end of the initial header line.
        """

        self.file_object.seek(0, os.SEEK_SET)
        head_line = self.file_object.readline().strip()
        parts = head_line.split(b'/')
        if len(parts) != 2:
            raise ValueError('Cannot extract CPHD version number from line {}'.format(head_line))
        cphd_version = parts[1].strip().decode('utf-8')
        self._cphd_version = cphd_version

    def _extract_header(self) -> None:
        """
        Extract the header from the file. The file object is assumed to be advanced
        to the header location. This will advance to the file object to the end of
        the header section.
        """

        if self.cphd_version.startswith('0.3'):
            self._cphd_header = CPHDHeader0_3.from_file_object(self._file_object)
        elif self.cphd_version.startswith('1.'):
            self._cphd_header = CPHDHeader1.from_file_object(self._file_object)
        else:
            raise ValueError(_unhandled_version_text.format(self.cphd_version))

    def _extract_cphd(self) -> None:
        """
        Extract and interpret the CPHD structure from the file.
        """

        xml = self.get_cphd_bytes()
        if self.cphd_version.startswith('0.3'):
            the_type = CPHDType0_3
        elif self.cphd_version.startswith('1.'):
            the_type = CPHDType1
        else:
            raise ValueError(_unhandled_version_text.format(self.cphd_version))

        self._cphd_meta = the_type.from_xml_string(xml)

    def get_cphd_bytes(self) -> bytes:
        """
        Extract the (uninterpreted) bytes representation of the CPHD structure.

        Returns
        -------
        bytes
        """

        header = self.cphd_header
        if header is None:
            raise ValueError('No cphd_header populated.')

        if self.cphd_version.startswith('0.3'):
            assert isinstance(header, CPHDHeader0_3)
            # extract the xml data
            self.file_object.seek(header.XML_BYTE_OFFSET, os.SEEK_SET)
            xml = self.file_object.read(header.XML_DATA_SIZE)
        elif self.cphd_version.startswith('1.'):
            assert isinstance(header, CPHDHeader1)
            # extract the xml data
            self.file_object.seek(header.XML_BLOCK_BYTE_OFFSET, os.SEEK_SET)
            xml = self.file_object.read(header.XML_BLOCK_SIZE)
        else:
            raise ValueError(_unhandled_version_text.format(self.cphd_version))
        return xml

    def close(self):
        if self._closed:
            return

        if self._close_after:
            if hasattr(self.file_object, 'close'):
                self.file_object.close()
        self._file_object = None
        self._closed = True

    def __del__(self):
        self.close()


def _validate_cphd_details(
        cphd_details: Union[str, CPHDDetails],
        version: Union[None, str, Sequence[str]] = None) -> CPHDDetails:
    """
    Validate the input argument.

    Parameters
    ----------
    cphd_details : str|CPHDDetails
    version : None|str|Sequence[str]

    Returns
    -------
    CPHDDetails

    Raises
    ------
    TypeError
        The input was neither path to a CPHD file nor a CPHDDetails instance
    ValueError
        The CPHD file was the incorrect (specified) version
    """

    if isinstance(cphd_details, str):
        cphd_details = CPHDDetails(cphd_details)

    if not isinstance(cphd_details, CPHDDetails):
        raise TypeError('cphd_details is required to be a file path to a CPHD file '
                        'or CPHDDetails, got type {}'.format(cphd_details))
    if version is not None:
        if isinstance(version, str) and not cphd_details.cphd_version.startswith(version):
            raise ValueError(
                'This CPHD file is required to be version {},\n\t'
                'got {}'.format(version, cphd_details.cphd_version))
        else:
            val = False
            for entry in version:
                if cphd_details.cphd_version.startswith(entry):
                    val = True
                    break
            if not val:
                raise ValueError(
                    'This CPHD file is required to be one of version {},\n\t'
                    'got {}'.format(version, cphd_details.cphd_version))

    return cphd_details


##########
# Reading


class CPHDReader(CPHDTypeReader):
    """
    The Abstract CPHD reader instance, which just selects the proper CPHD reader
    class based on the CPHD version. Note that there is no __init__ method for
    this class, and it would be skipped regardless. Ensure that you make a direct
    call to the BaseReader.__init__() method when extending this class.

    **Updated in version 1.3.0** for reading changes.
    """

    __slots__ = ('_cphd_details', )

    def __new__(cls, *args, **kwargs):
        if len(args) == 0:
            raise ValueError(
                'The first argument of the constructor is required to be a file_path '
                'or CPHDDetails instance.')
        if is_file_like(args[0]):
            raise ValueError('File like object input not supported for CPHD reading at this time.')
        cphd_details = _validate_cphd_details(args[0])

        if cphd_details.cphd_version.startswith('0.3'):
            return object.__new__(CPHDReader0_3)
        elif cphd_details.cphd_version.startswith('1.'):
            return object.__new__(CPHDReader1)
        else:
            raise ValueError('Got unhandled CPHD version {}'.format(cphd_details.cphd_version))

    @property
    def cphd_details(self) -> CPHDDetails:
        """
        CPHDDetails: The cphd details object.
        """

        return self._cphd_details

    @property
    def cphd_version(self) -> str:
        """
        str: The CPHD version.
        """

        return self.cphd_details.cphd_version

    @property
    def cphd_header(self) -> Union[CPHDHeader1, CPHDHeader0_3]:
        """
        CPHDHeader1|CPHDHeader0_3: The CPHD header object, which is version dependent.
        """

        return self.cphd_details.cphd_header

    @property
    def file_name(self) -> str:
        return self.cphd_details.file_name

    def read_pvp_variable(
            self,
            variable: str,
            index: Union[int, str],
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> Optional[numpy.ndarray]:
        raise NotImplementedError

    def read_pvp_array(
            self,
            index: Union[int, str],
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> numpy.ndarray:
        raise NotImplementedError

    def read_pvp_block(self) -> Dict[Union[int, str], numpy.ndarray]:
        raise NotImplementedError

    def read_signal_block(self) -> Dict[Union[int, str], numpy.ndarray]:
        raise NotImplementedError

    def read_signal_block_raw(self) -> Dict[Union[int, str], numpy.ndarray]:
        raise NotImplementedError

    def close(self):
        CPHDTypeReader.close(self)
        if hasattr(self, '_cphd_details'):
            if hasattr(self._cphd_details, 'close'):
                self._cphd_details.close()
            del self._cphd_details


class CPHDReader1(CPHDReader):
    """
    The CPHD version 1 reader.

    **Updated in version 1.3.0** for reading changes.
    """
    _allowed_versions = ('1.0', '1.1')

    def __new__(cls, *args, **kwargs):
        # we must override here, to avoid recursion with
        # the CPHDReader parent
        return object.__new__(cls)

    def __init__(self, cphd_details: Union[str, CPHDDetails]):
        """

        Parameters
        ----------
        cphd_details : str|CPHDDetails
        """

        self._channel_map = None  # type: Union[None, Dict[str, int]]
        self._pvp_memmap = None  # type: Union[None, Dict[str, numpy.ndarray]]
        self._support_array_memmap = None  # type: Union[None, Dict[str, numpy.ndarray]]
        self._cphd_details = _validate_cphd_details(cphd_details, version=self._allowed_versions)

        CPHDTypeReader.__init__(self, None, self._cphd_details.cphd_meta)
        # set data segments after setting up the pvp information, because
        #   we need the AmpSf to set up the format function for the data segment
        self._create_pvp_memmaps()
        self._create_support_array_memmaps()

        data_segments = self._create_data_segments()
        BaseReader.__init__(self, data_segments, reader_type='CPHD')

    @property
    def cphd_meta(self) -> CPHDType1:
        """
        CPHDType1: The CPHD structure.
        """

        return self._cphd_meta

    @property
    def cphd_header(self) -> CPHDHeader1:
        """
        CPHDHeader1: The CPHD header object.
        """

        return self.cphd_details.cphd_header

    def _create_data_segments(self) -> List[DataSegment]:
        """
        Helper method for creating the various signal data segments.

        Returns
        -------
        List[DataSegment]
        """

        data_segments = []

        data = self.cphd_meta.Data
        sample_type = data.SignalArrayFormat

        compressed = data.SignalCompressionID
        if compressed is not None:
            raw_dtype = numpy.dtype('B')
        elif sample_type == "CF8":
            raw_dtype = numpy.dtype('>f4')
        elif sample_type == "CI4":
            raw_dtype = numpy.dtype('>i2')
        elif sample_type == "CI2":
            raw_dtype = numpy.dtype('>i1')
        else:
            raise ValueError('Got unhandled signal array format {}'.format(sample_type))

        block_offset = self.cphd_header.SIGNAL_BLOCK_BYTE_OFFSET
        for entry in data.Channels:
            amp_sf = self.read_pvp_variable('AmpSF', entry.Identifier)
            if compressed:
                raw_shape = (entry.CompressedSignalSize,)
                formatted_shape = raw_shape
                formatted_dtype = 'byte'
                format_function = None
            else:
                raw_shape = (entry.NumVectors, entry.NumSamples, 2)
                formatted_shape = raw_shape[:2]
                formatted_dtype = 'complex64'
                format_function = AmpScalingFunction(raw_dtype, amplitude_scaling=amp_sf)

            data_offset = entry.SignalArrayByteOffset
            data_segments.append(
                NumpyMemmapSegment(
                    self.cphd_details.file_object, block_offset+data_offset,
                    raw_dtype, raw_shape, formatted_dtype=formatted_dtype, formatted_shape = formatted_shape,
                    format_function=format_function, close_file=False))
        return data_segments

    def _create_pvp_memmaps(self) -> None:
        """
        Helper method which creates the pvp mem_maps.

        Returns
        -------
        None
        """

        self._pvp_memmap = None
        if self.cphd_meta.Data.Channels is None:
            logger.error('No Data.Channels defined.')
            return
        if self.cphd_meta.PVP is None:
            logger.error('No PVP object defined.')
            return

        pvp_dtype = self.cphd_meta.PVP.get_vector_dtype()
        self._pvp_memmap = OrderedDict()
        self._channel_map = OrderedDict()
        for i, entry in enumerate(self.cphd_meta.Data.Channels):
            self._channel_map[entry.Identifier] = i
            offset = self.cphd_header.PVP_BLOCK_BYTE_OFFSET + entry.PVPArrayByteOffset
            shape = (entry.NumVectors, )
            self._pvp_memmap[entry.Identifier] = numpy.memmap(
                self.cphd_details.file_name, dtype=pvp_dtype, mode='r', offset=offset, shape=shape)

    def _create_support_array_memmaps(self) -> None:
        """
        Helper method which creates the support array mem_maps.

        Returns
        -------
        None
        """

        if self.cphd_meta.Data.SupportArrays is None:
            self._support_array_memmap = None
            return

        self._support_array_memmap = OrderedDict()
        for i, entry in enumerate(self.cphd_meta.Data.SupportArrays):
            # extract the support array metadata details
            details = self.cphd_meta.SupportArray.find_support_array(entry.Identifier)
            # determine array byte offset
            offset = self.cphd_header.SUPPORT_BLOCK_BYTE_OFFSET + entry.ArrayByteOffset
            # determine numpy dtype and depth of array
            dtype, depth = details.get_numpy_format()
            # set up the numpy memory map
            shape = (entry.NumRows, entry.NumCols) if depth == 1 else (entry.NumRows, entry.NumCols, depth)
            self._support_array_memmap[entry.Identifier] = numpy.memmap(
                self.cphd_details.file_name, dtype=dtype, mode='r', offset=offset, shape=shape)

    def _validate_index(self, index: Union[int, str]) -> int:
        """
        Get corresponding integer index for CPHD channel.

        Parameters
        ----------
        index : int|str

        Returns
        -------
        int
        """

        cphd_meta = self.cphd_details.cphd_meta

        if isinstance(index, str):
            if index in self._channel_map:
                return self._channel_map[index]
            else:
                raise KeyError(_missing_channel_identifier_text.format(index))
        else:
            int_index = int(index)
            if not (0 <= int_index < cphd_meta.Data.NumCPHDChannels):
                raise ValueError(_index_range_text.format(cphd_meta.Data.NumCPHDChannels))
            return int_index

    def _validate_index_key(self, index: Union[int, str]) -> str:
        """
        Gets the corresponding identifier for the CPHD channel.

        Parameters
        ----------
        index : int|str

        Returns
        -------
        str
        """

        cphd_meta = self.cphd_details.cphd_meta

        if isinstance(index, str):
            if index in self._channel_map:
                return index
            else:
                raise KeyError(_missing_channel_identifier_text.format(index))
        else:
            int_index = int(index)
            if not (0 <= int_index < cphd_meta.Data.NumCPHDChannels):
                raise ValueError(_index_range_text.format(cphd_meta.Data.NumCPHDChannels))
            return cphd_meta.Data.Channels[int_index].Identifier

    def read_support_array(
            self,
            index: Union[int, str],
            *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]]) -> numpy.ndarray:
        # find the support array identifier
        if isinstance(index, int):
            the_entry = self.cphd_meta.Data.SupportArrays[index]
            index = the_entry.Identifier
        if not isinstance(index, str):
            raise TypeError('Got unexpected type {} for identifier'.format(type(index)))

        the_memmap = self._support_array_memmap[index]

        if len(ranges) == 0:
            return numpy.copy(the_memmap[:])

        # noinspection PyTypeChecker
        subscript = verify_subscript(ranges, the_memmap.shape)
        return numpy.copy(the_memmap[subscript])

    def read_support_block(self) -> Dict[str, numpy.ndarray]:
        if self.cphd_meta.Data.SupportArrays:
            return {
                sa.Identifier: self.read_support_array(sa.Identifier)
                for sa in self.cphd_meta.Data.SupportArrays}
        else:
            return {}

    def read_pvp_variable(
            self,
            variable: str,
            index: Union[int, str],
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> Optional[numpy.ndarray]:
        index_key = self._validate_index_key(index)
        the_memmap = self._pvp_memmap[index_key]
        the_slice = verify_slice(the_range, the_memmap.shape[0])
        if variable in the_memmap.dtype.fields:
            return numpy.copy(the_memmap[variable][the_slice])
        else:
            return None

    def read_pvp_array(
            self,
            index: Union[int, str],
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> numpy.ndarray:
        index_key = self._validate_index_key(index)
        the_memmap = self._pvp_memmap[index_key]
        the_slice = verify_slice(the_range, the_memmap.shape[0])
        return numpy.copy(the_memmap[the_slice])

    def read_pvp_block(self) -> Dict[str, numpy.ndarray]:
        return {chan.Identifier: self.read_pvp_array(chan.Identifier)
                for chan in self.cphd_meta.Data.Channels}

    def read_signal_block(self) -> Dict[str, numpy.ndarray]:
        return {chan.Identifier: numpy.copy(self.read(index=chan.Identifier))
                for chan in self.cphd_meta.Data.Channels}

    def read_signal_block_raw(self) -> Dict[Union[int, str], numpy.ndarray]:
        return {chan.Identifier: numpy.copy(self.read_raw(index=chan.Identifier))
                for chan in self.cphd_meta.Data.Channels}

    def read_chip(
            self,
            *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]],
            index: Union[int, str] = 0,
            squeeze: bool = True) -> numpy.ndarray:
        """
        This is identical to :meth:`read`, and presented for backwards compatibility.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
        index : int|str
        squeeze : bool

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        :meth:`read`.
        """

        return self.__call__(*ranges, index=index, raw=False, squeeze=squeeze)

    def read(
            self,
            *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]],
            index: Union[int, str] = 0,
            squeeze: bool = True) -> numpy.ndarray:
        """
        Read formatted data from the given data segment. Note this is an alias to the
        :meth:`__call__` called as
        :code:`reader(*ranges, index=index, raw=False, squeeze=squeeze)`.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
            The slice definition appropriate for `data_segment[index].read()` usage.
        index : int|str
            The data_segment index or channel identifier.
        squeeze : bool
            Squeeze length 1 dimensions out of the shape of the return array?

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.read`.
        """

        return self.__call__(*ranges, index=index, raw=False, squeeze=squeeze)

    def read_raw(
            self,
            *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]],
            index: Union[int, str] = 0,
            squeeze: bool = True) -> numpy.ndarray:
        """
        Read raw data from the given data segment. Note this is an alias to the
        :meth:`__call__` called as
        :code:`reader(*ranges, index=index, raw=True, squeeze=squeeze)`.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
            The slice definition appropriate for `data_segment[index].read()` usage.
        index : int|str
            The data_segment index or cphd channel identifier.
        squeeze : bool
            Squeeze length 1 dimensions out of the shape of the return array?

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.read_raw`.
        """

        return self.__call__(*ranges, index=index, raw=True, squeeze=squeeze)

    def __call__(
            self,
            *ranges: Sequence[Union[None, int, slice]],
            index: int = 0,
            raw: bool = False,
            squeeze: bool = True) -> numpy.ndarray:
        index = self._validate_index(index)
        return BaseReader.__call__(self, *ranges, index=index, raw=raw, squeeze=squeeze)


class CPHDReader0_3(CPHDReader):
    """
    The CPHD version 0.3 reader.

    **Updated in version 1.3.0** for reading changes.
    """

    def __new__(cls, *args, **kwargs):
        # we must override here, to avoid recursion with
        # the CPHDReader parent
        return object.__new__(cls)

    def __init__(self, cphd_details: Union[str, CPHDDetails]):
        """

        Parameters
        ----------
        cphd_details : str|CPHDDetails
        """

        self._cphd_details = _validate_cphd_details(cphd_details, version='0.3')
        CPHDTypeReader.__init__(self, None, self._cphd_details.cphd_meta)
        self._create_pvp_memmaps()

        data_segments = self._create_data_segment()
        BaseReader.__init__(self, data_segments, reader_type="CPHD")

    @property
    def cphd_meta(self) -> CPHDType0_3:
        """
        CPHDType0_3: The CPHD structure, which is version dependent.
        """

        return self._cphd_meta

    @property
    def cphd_header(self) -> CPHDHeader0_3:
        """
        CPHDHeader0_3: The CPHD header object.
        """

        return self.cphd_details.cphd_header

    def _validate_index(self, index: int) -> int:
        """
        Validate integer index value for CPHD channel.

        Parameters
        ----------
        index : int

        Returns
        -------
        int
        """

        int_index = int(index)
        if not (0 <= int_index < self.cphd_meta.Data.NumCPHDChannels):
            raise ValueError(_index_range_text.format(self.cphd_meta.Data.NumCPHDChannels))
        return int_index

    def _create_data_segment(self) -> List[DataSegment]:
        data_segments = []

        data = self.cphd_meta.Data
        sample_type = data.SampleType
        if sample_type == "RE32F_IM32F":
            raw_dtype = numpy.dtype('>f4')
        elif sample_type == "RE16I_IM16I":
            raw_dtype = numpy.dtype('>i2')
        elif sample_type == "RE08I_IM08I":
            raw_dtype = numpy.dtype('>i1')
        else:
            raise ValueError('Got unhandled sample type {}'.format(sample_type))

        data_offset = self.cphd_header.CPHD_BYTE_OFFSET
        for index, entry in enumerate(data.ArraySize):
            amp_sf = self.read_pvp_variable('AmpSF', index)
            format_function = AmpScalingFunction(raw_dtype, amplitude_scaling=amp_sf)
            raw_shape = (entry.NumVectors, entry.NumSamples, 2)
            data_segments.append(
                NumpyMemmapSegment(
                    self.cphd_details.file_object, data_offset,
                    raw_dtype, raw_shape, formatted_dtype='complex64', formatted_shape=raw_shape[:2],
                    format_function=format_function, close_file=False))
            data_offset += raw_shape[0]*raw_shape[1]*2*raw_dtype.itemsize
        return data_segments

    def _create_pvp_memmaps(self) -> None:
        """
        Helper method which creates the pvp mem_maps.

        Returns
        -------
        None
        """

        self._pvp_memmap = None
        pvp_dtype = self.cphd_meta.VectorParameters.get_vector_dtype()
        self._pvp_memmap = []
        for i, entry in enumerate(self.cphd_meta.Data.ArraySize):
            offset = self.cphd_header.VB_BYTE_OFFSET + self.cphd_meta.Data.NumBytesVBP*i
            shape = (entry.NumVectors, )
            self._pvp_memmap.append(
                numpy.memmap(
                    self.cphd_details.file_name, dtype=pvp_dtype, mode='r', offset=offset, shape=shape))

    def read_pvp_variable(
            self,
            variable: str,
            index: int,
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> Optional[numpy.ndarray]:
        int_index = self._validate_index(index)
        the_memmap = self._pvp_memmap[int_index]
        the_slice = verify_slice(the_range, the_memmap.shape[0])
        if variable in the_memmap.dtype.fields:
            return numpy.copy(the_memmap[variable][the_slice])
        else:
            return None

    def read_pvp_array(
            self,
            index: int,
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> numpy.ndarray:
        int_index = self._validate_index(index)
        the_memmap = self._pvp_memmap[int_index]
        the_slice = verify_slice(the_range, the_memmap.shape[0])
        return numpy.copy(the_memmap[the_slice])

    def read_pvp_block(self) -> Dict[int, numpy.ndarray]:
        """
        Reads the entirety of the PVP block(s).

        Returns
        -------
        Dict[int, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the PVP arrays.
        """

        return {chan: self.read_pvp_array(chan) for chan in range(self.cphd_meta.Data.NumCPHDChannels)}

    def read_signal_block(self) -> Dict[int, numpy.ndarray]:
        return {chan: self.read(index=chan) for chan in range(self.cphd_meta.Data.NumCPHDChannels)}

    def read_signal_block_raw(self) -> Dict[int, numpy.ndarray]:
        return {chan: self.read_raw(index=chan) for chan in range(self.cphd_meta.Data.NumCPHDChannels)}

    def __call__(
            self,
            *ranges: Sequence[Union[None, int, slice]],
            index: int = 0,
            raw: bool = False,
            squeeze: bool = True) -> numpy.ndarray:
        index = self._validate_index(index)
        return BaseReader.__call__(self, *ranges, index=index, raw=raw, squeeze=squeeze)


def is_a(file_name: str) -> Optional[CPHDReader]:
    """
    Tests whether a given file_name corresponds to a CPHD file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    CPHDReader|None
        Appropriate `CPHDTypeReader` instance if CPHD file, `None` otherwise
    """

    try:
        cphd_details = CPHDDetails(file_name)
        logger.info('File {} is determined to be a CPHD version {} file.'.format(file_name, cphd_details.cphd_version))
        return CPHDReader(cphd_details)
    except SarpyIOError:
        # we don't want to catch parsing errors, for now?
        return None


###########
# Writing

class ElementDetails(object):
    __slots__ = (
        '_item_offset', '_item_bytes', '_item_written')

    def __init__(self, item_offset: int, item_bytes: Optional[bytes] = None):
        self._item_offset = None
        self._item_bytes = None
        self._item_written = False

        self.item_offset = item_offset
        self.item_bytes = item_bytes

    @property
    def item_offset(self) -> Optional[int]:
        """
        int: The item offset.
        """

        return self._item_offset

    @item_offset.setter
    def item_offset(self, value: int) -> None:
        value = int(value)
        if self._item_offset is not None and self._item_offset != value:
            raise ValueError("item_offset is read only after being initially defined.")
        self._item_offset = value

    @property
    def item_bytes(self) -> Optional[bytes]:
        """
        None|bytes: The item bytes.
        """

        return self._item_bytes

    @item_bytes.setter
    def item_bytes(self, value: bytes) -> None:
        if self._item_bytes is not None:
            raise ValueError("item_bytes is read only after being initially defined.")
        if value is None:
            self._item_bytes = None
            return

        if not isinstance(value, bytes):
            raise TypeError('item_bytes must be of type bytes')
        self._item_bytes = value

    @property
    def item_written(self) -> bool:
        """
        bool: Has the item been written?
        """

        return self._item_written

    @item_written.setter
    def item_written(self, value: bool):
        value = bool(value)
        if self._item_written and not value:
            raise ValueError(
                'item_written has already been set to True,\n\t'
                'it cannot be reverted to False')
        self._item_written = value

    def write_item(self, file_object: BinaryIO) -> None:
        """
        Write the item bytes (if populated), at its specified offset, to the
        file. This requires that the subheader has previously be written. If
        writing occurs, the file location will be advanced to the end of the item
        location.

        Parameters
        ----------
        file_object : BinaryIO

        Returns
        -------
        None
        """

        if self.item_written:
            return

        if self.item_offset is None:
            return  # nothing to be done

        if self.item_bytes is None:
            return  # nothing to be done

        file_object.seek(self.item_offset, os.SEEK_SET)
        file_object.write(self.item_bytes)
        self.item_written = True


class CPHDWritingDetails(object):
    __slots__ = (
        '_header', '_header_written', '_meta',
        '_channel_map', '_support_map',
        '_pvp_details', '_support_details', '_signal_details')

    def __init__(self, meta: CPHDType1, check_older_version: bool = False):

        self._header = None
        self._header_written = False
        self._meta = None
        self._channel_map = {}
        self._support_map = {}
        self._pvp_details = None
        self._support_details = None
        self._signal_details = None

        self.meta = meta
        self._set_header(check_older_version)

        # initialize the information for the pvp, support, and signal details
        self._populate_pvp_details()
        self._populate_support_details()
        self._populate_signal_details()

    @property
    def header(self) -> CPHDHeader1:
        return self._header

    def _set_header(self, check_older_version: bool):
        if check_older_version:
            use_version_tuple = self.meta.version_required()
        else:
            use_version_tuple = get_default_tuple()
        use_version_string = '{}.{}.{}'.format(*use_version_tuple)
        self._header = self.meta.make_file_header(use_version=use_version_string)

    @property
    def use_version(self) -> str:
        return self.header.use_version

    @property
    def meta(self) -> CPHDType1:
        """
        CPHDType1: The metadata
        """

        return self._meta

    @meta.setter
    def meta(self, value):
        if self._meta is not None:
            raise ValueError('meta is read only once initialized.')
        if not isinstance(value, CPHDType1):
            raise TypeError('meta must be of type {}'.format(CPHDType1))
        self._meta = value

    def _populate_pvp_details(self) -> None:
        if self._pvp_details is not None:
            raise ValueError('pvp_details can not be initialized again')
        pvp_details = []
        for i, entry in enumerate(self.meta.Data.Channels):
            self._channel_map[entry.Identifier] = i
            offset = self.header.PVP_BLOCK_BYTE_OFFSET + entry.PVPArrayByteOffset
            pvp_details.append(ElementDetails(offset))
        self._pvp_details = tuple(pvp_details)

    def _populate_support_details(self) -> None:
        if self._support_details is not None:
            raise ValueError('support_details can not be initialized again')

        if self.meta.Data.SupportArrays is None:
            self._signal_details = None
            return

        support_details = []
        for i, entry in enumerate(self.meta.Data.SupportArrays):
            self._support_map[entry.Identifier] = i
            offset = self.header.SUPPORT_BLOCK_BYTE_OFFSET + entry.ArrayByteOffset
            support_details.append(ElementDetails(offset))
        self._support_details = tuple(support_details)

    def _populate_signal_details(self) -> None:
        if self._signal_details is not None:
            raise ValueError('signal_details can not be initialized again')

        signal_details = []
        for i, entry in enumerate(self.meta.Data.Channels):
            offset = self.header.SIGNAL_BLOCK_BYTE_OFFSET + entry.SignalArrayByteOffset
            signal_details.append(ElementDetails(offset))
        self._signal_details = tuple(signal_details)

    @property
    def pvp_details(self) -> Optional[Tuple[ElementDetails, ...]]:
        return self._pvp_details

    @property
    def support_details(self) -> Optional[Tuple[ElementDetails, ...]]:
        return self._support_details

    @property
    def signal_details(self) -> Optional[Tuple[ElementDetails, ...]]:
        return self._signal_details

    @property
    def channel_map(self) -> Dict[str, int]:
        return self._channel_map

    @property
    def support_map(self) -> Optional[Dict[str, int]]:
        return self._support_map

    def _write_items(
            self,
            details: Optional[Sequence[ElementDetails]],
            file_object: BinaryIO) -> None:
        if details is None:
            return
        for index, entry in enumerate(details):
            entry.write_item(file_object)

    def _verify_item_written(
            self,
            details: Optional[Sequence[ElementDetails]],
            name: str) -> None:
        if details is None:
            return

        for index, entry in enumerate(details):
            if not entry.item_written:
                logger.error('{} data at index {} not written'.format(name, index))

    def write_header(
            self,
            file_object: BinaryIO,
            overwrite: bool = False) -> None:
        """
        Write the header.The file object will be advanced to the end of the
        block, if writing occurs.

        Parameters
        ----------
        file_object : BinaryIO
        overwrite : bool
            Overwrite, if previously written?

        Returns
        -------
        None
        """

        if self._header_written and not overwrite:
            return

        file_object.write(self.header.to_string().encode())
        file_object.write(CPHD_SECTION_TERMINATOR)
        # write xml
        file_object.seek(self.header.XML_BLOCK_BYTE_OFFSET, os.SEEK_SET)
        file_object.write(self.meta.to_xml_bytes(urn=get_namespace(self.use_version)))
        file_object.write(CPHD_SECTION_TERMINATOR)
        self._header_written = True

    def write_all_populated_items(self, file_object: BinaryIO) -> None:
        """
        Write everything populated. This assumes that the header will start at the
        beginning (position 0) of the file-like object.

        Parameters
        ----------
        file_object : BinaryIO

        Returns
        -------
        None
        """

        self.write_header(file_object, overwrite=False)
        self._write_items(self.pvp_details, file_object)
        self._write_items(self.support_details, file_object)
        self._write_items(self.signal_details, file_object)

    def verify_all_written(self) -> None:
        if not self._header_written:
            logger.error('header not written')

        self._verify_item_written(self.pvp_details, 'pvp')
        self._verify_item_written(self.support_details, 'support')
        self._verify_item_written(self.signal_details, 'signal')


class CPHDWriter1(BaseWriter):
    """
    The CPHD version 1 writer.

    **Updated in version 1.3.0** for writing changes.
    """
    _writing_details_type = CPHDWritingDetails

    __slots__ = (
        '_file_name', '_file_object', '_in_memory', '_writing_details',
        '_pvp_memmaps', '_support_memmaps', '_signal_data_segments',
        '_can_write_regular_data')

    def __init__(
            self,
            file_object: Union[str, BinaryIO],
            meta: Optional[CPHDType1] = None,
            writing_details: Optional[CPHDWritingDetails] = None,
            check_older_version: bool = False,
            check_existence: bool = True):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
        meta : None|CPHDType1
        writing_details : None|CPHDWritingDetails
        check_older_version : bool
            Try to create an older version CPHD for compliance with other
            NGA applications
        check_existence : bool
            Should we check if the given file already exists, and raises an exception if so?
        """

        self._writing_details = None

        if isinstance(file_object, str):
            if check_existence and os.path.exists(file_object):
                raise SarpyIOError(
                    'Given file {} already exists, and a new CPHD file cannot be created here.'.format(file_object))
            file_object = open(file_object, 'wb')

        if not is_file_like(file_object):
            raise ValueError('file_object requires a file path or BinaryIO object')

        self._file_object = file_object
        if is_real_file(file_object):
            self._file_name = file_object.name
            self._in_memory = False
        else:
            self._file_name = None
            self._in_memory = True

        if meta is None and writing_details is None:
            raise ValueError('One of meta or writing_details must be provided.')
        if writing_details is None:
            writing_details = self._writing_details_type(meta, check_older_version=check_older_version)
        self.writing_details = writing_details

        self._pvp_memmaps = None  # type: Optional[Dict[str, numpy.ndarray]]
        self._support_memmaps = None  # type: Optional[Dict[str, numpy.ndarray]]
        self._signal_data_segments = None  # type: Optional[Dict[str, DataSegment]]
        self._can_write_regular_data = None  # type: Optional[Dict[str, bool]]
        self._closed = False

        data_segment = self._initialize_data()
        BaseWriter.__init__(self, data_segment)

    @property
    def writing_details(self) -> CPHDWritingDetails:
        return self._writing_details

    @writing_details.setter
    def writing_details(self, value):
        if self._writing_details is not None:
            raise ValueError('writing_details is read-only')
        if not isinstance(value, CPHDWritingDetails):
            raise TypeError('writing_details must be of type {}'.format(CPHDWritingDetails))
        self._writing_details = value

    @property
    def file_name(self) -> Optional[str]:
        return self._file_name

    @property
    def meta(self) -> CPHDType1:
        """
        CPHDType1: The metadata
        """

        return self.writing_details.meta

    @staticmethod
    def _verify_dtype(
            obs_dtype: numpy.dtype,
            exp_dtype: numpy.dtype,
            purpose: str) -> None:
        """
        This is a helper function for comparing two structured array dtypes.

        Parameters
        ----------
        obs_dtype : numpy.dtype
        exp_dtype : numpy.dtype
        purpose : str
        """

        if obs_dtype.fields is None or exp_dtype.fields is None:
            raise ValueError('structure array dtype required.')

        observed_dtype = sorted(
            [(field, dtype_info) for field, dtype_info in obs_dtype.fields.items()],
            key=lambda x: x[1][1])
        expected_dtype = sorted(
            [(field, dtype_info) for field, dtype_info in exp_dtype.fields.items()],
            key=lambda x: x[1][1])
        if len(observed_dtype) != len(expected_dtype):
            raise ValueError('Observed dtype for {} does not match the expected dtype.'.format(purpose))
        for obs_entry, exp_entry in zip(observed_dtype, expected_dtype):
            if obs_entry[1][1] != exp_entry[1][1]:
                raise ValueError(
                    'Observed dtype for {} does not match the expected dtype\nobserved {}\nexpected {}.'.format(
                        purpose, observed_dtype, expected_dtype))
            if obs_entry[0] != exp_entry[0]:
                logger.warning(
                    'Got mismatched field names (observed {}, expected {}) for {}.'.format(
                        obs_entry[0], exp_entry[0], purpose))

    def _validate_channel_index(self, index: Union[int, str]) -> int:
        """
        Get corresponding integer index for CPHD channel.

        Parameters
        ----------
        index : int|str

        Returns
        -------
        int
        """

        if isinstance(index, str):
            if index in self.writing_details.channel_map:
                return self.writing_details.channel_map[index]
            else:
                raise KeyError(_missing_channel_identifier_text.format(index))
        else:
            int_index = int(index)
            if not (0 <= int_index < self.meta.Data.NumCPHDChannels):
                raise ValueError(_index_range_text.format(self.meta.Data.NumCPHDChannels))
            return int_index

    def _validate_channel_key(self, index: Union[int, str]) -> str:
        """
        Gets the corresponding identifier for the CPHD channel.

        Parameters
        ----------
        index : int|str

        Returns
        -------
        str
        """

        if isinstance(index, str):
            if index in self.writing_details.channel_map:
                return index
            else:
                raise KeyError(_missing_channel_identifier_text.format(index))
        else:
            int_index = int(index)
            if not (0 <= int_index < self.meta.Data.NumCPHDChannels):
                raise ValueError(_index_range_text.format(self.meta.Data.NumCPHDChannels))
            return self.meta.Data.Channels[int_index].Identifier

    def _validate_support_index(self, index: Union[int, str]) -> int:
        """
        Get corresponding integer index for support array.

        Parameters
        ----------
        index : int|str

        Returns
        -------
        int
        """

        if isinstance(index, str):
            if index in self.writing_details.support_map:
                return self.writing_details.support_map[index]
            else:
                raise KeyError('Cannot find support array for identifier {}'.format(index))
        else:
            int_index = int(index)
            if not (0 <= int_index < len(self.meta.Data.SupportArrays)):
                raise ValueError(_index_range_text.format(len(self.meta.Data.SupportArrays)))
            return int_index

    def _validate_support_key(self, index: Union[int, str]) -> str:
        """
        Gets the corresponding identifier for the support array.

        Parameters
        ----------
        index : int|str

        Returns
        -------
        str
        """

        if isinstance(index, str):
            if index in self.writing_details.support_map:
                return index
            else:
                raise KeyError('Cannot find support array for identifier {}'.format(index))
        else:
            int_index = int(index)
            if not (0 <= int_index < len(self.meta.Data.SupportArrays)):
                raise ValueError(_index_range_text.format(len(self.meta.Data.SupportArrays)))
            return self.meta.Data.SupportArrays[int_index].Identifier

    def _initialize_data(self) -> List[DataSegment]:
        self._pvp_memmaps = {}
        # set up the PVP memmaps
        pvp_dtype = self.meta.PVP.get_vector_dtype()
        for i, entry in enumerate(self.meta.Data.Channels):
            # create the pvp mem map
            offset = self.writing_details.pvp_details[i].item_offset
            shape = (entry.NumVectors, )
            if self._in_memory:
                self._pvp_memmaps[entry.Identifier] = numpy.empty(shape, dtype=pvp_dtype)
            else:
                self._pvp_memmaps[entry.Identifier] = numpy.memmap(
                    self._file_name, dtype=pvp_dtype, mode='r+', offset=offset, shape=shape)

        self._support_memmaps = {}
        if self.meta.Data.SupportArrays is not None:
            for i, entry in enumerate(self.meta.Data.SupportArrays):
                # extract the support array metadata details
                details = self.meta.SupportArray.find_support_array(entry.Identifier)
                offset = self.writing_details.support_details[i].item_offset
                # determine numpy dtype and depth of array
                dtype, depth = details.get_numpy_format()
                # set up the numpy memory map
                shape = (entry.NumRows, entry.NumCols) if depth == 1 else (entry.NumRows, entry.NumCols, depth)
                if self._in_memory:
                    self._support_memmaps[entry.Identifier] = numpy.empty(shape, dtype=dtype)
                else:
                    self._support_memmaps[entry.Identifier] = numpy.memmap(
                        self._file_name, dtype=dtype, mode='r+', offset=offset, shape=shape)

        # set up the signal data_segment (this is used for formatting issues)
        no_amp_sf = (self.meta.PVP.AmpSF is None)
        self._signal_data_segments = {}
        self._can_write_regular_data = {}
        signal_data_segments = []
        signal_array_format = self.meta.Data.SignalArrayFormat
        compressed = getattr(self.meta.Data, 'SignalCompressionID', None)
        if compressed is not None:
            signal_dtype = numpy.dtype('B')
        elif signal_array_format == 'CI2':
            signal_dtype = numpy.dtype('>i1')
        elif signal_array_format == 'CI4':
            signal_dtype = numpy.dtype('>i2')
        elif signal_array_format == 'CF8':
            signal_dtype = numpy.dtype('>f4')
        else:
            raise ValueError('Got unhandled SignalArrayFormat {}'.format(signal_array_format))
        for i, entry in enumerate(self.meta.Data.Channels):
            self._can_write_regular_data[entry.Identifier] = no_amp_sf or compressed
            if compressed:
                raw_shape = (entry.CompressedSignalSize,)
                formatted_shape = raw_shape
                formatted_dtype = 'byte'
                format_function = None
            else:
                raw_shape = (entry.NumVectors, entry.NumSamples, 2)
                formatted_shape = raw_shape[:2]
                formatted_dtype = 'complex64'
                format_function = AmpScalingFunction(signal_dtype)

            offset = self.writing_details.signal_details[i].item_offset
            if self._in_memory:
                underlying_array = numpy.full(raw_shape, 0, dtype=signal_dtype)
                data_segment = NumpyArraySegment(
                    underlying_array, formatted_dtype, formatted_shape=formatted_shape,
                    format_function=format_function, mode='w')
            else:
                data_segment = NumpyMemmapSegment(
                    self._file_object.name, offset, signal_dtype, raw_shape,
                    formatted_dtype=formatted_dtype, formatted_shape=formatted_shape,
                    format_function=format_function, mode='w', close_file=False)
            signal_data_segments.append(data_segment)

            self._signal_data_segments[entry.Identifier] = data_segment
        return signal_data_segments

    def write_support_array(self,
                            identifier: Union[int, str],
                            data: numpy.ndarray) -> None:
        """
        Write support array data to the file.

        Parameters
        ----------
        identifier : int|str
        data : numpy.ndarray
        """

        self._validate_closed()

        int_index = self._validate_support_index(identifier)
        identifier = self._validate_support_key(identifier)

        out_array = self._support_memmaps[identifier]
        if data.shape != out_array.shape:
            raise ValueError(
                'Support data shape {} is not compatible with\n\t'
                'that provided in metadata {}'.format(data.shape, out_array.shape))

        # write the data
        out_array[:] = data
        # mark it as written
        details = self.writing_details.support_details[int_index]
        if self._in_memory:
            # TODO: we can delete the memmap now?
            details.item_bytes = out_array.tobytes()
        else:
            details.item_written = True

    def write_pvp_array(self,
                        identifier: Union[int, str],
                        data: numpy.ndarray) -> None:
        """
        Write the PVP array data to the file.

        Parameters
        ----------
        identifier : int|str
        data : numpy.ndarray
        """

        self._validate_closed()

        def validate_dtype():
            self._verify_dtype(data.dtype, self._pvp_memmaps[identifier].dtype, 'PVP channel {}'.format(identifier))

        int_index = self._validate_channel_index(identifier)
        identifier = self._validate_channel_key(identifier)
        entry = self.meta.Data.Channels[int_index]
        validate_dtype()

        if data.ndim != 1:
            raise ValueError('Provided data is required to be one dimensional')
        if data.shape[0] != entry.NumVectors:
            raise ValueError('Provided data must have size determined by NumVectors')

        if self.meta.PVP.AmpSF is not None and self.meta.Data.SignalCompressionID is None:
            amp_sf = numpy.copy(data['AmpSF'][:])
            # noinspection PyUnresolvedReferences
            self._signal_data_segments[identifier].format_function.set_amplitude_scaling(amp_sf)
            self._can_write_regular_data[identifier] = True

        # write the data
        self._pvp_memmaps[identifier][list(data.dtype.names)] = data[list(data.dtype.names)]
        # mark it as written
        details = self.writing_details.pvp_details[int_index]
        if self._in_memory:
            # TODO: we can likely delete the memmap now?
            details.item_bytes = self._pvp_memmaps[identifier].tobytes()
        else:
            details.item_written = True

    def write_support_block(self, support_block: Dict[Union[int, str], numpy.ndarray]) -> None:
        """
        Write support block to the file.

        Parameters
        ----------
        support_block: dict
            Dictionary of `numpy.ndarray` containing the support arrays.
        """

        expected_support_ids = {s.Identifier for s in self.meta.Data.SupportArrays}
        assert expected_support_ids == set(support_block), 'support_block keys do not match those in meta'
        for identifier, array in support_block.items():
            self.write_support_array(identifier, array)

    def write_pvp_block(self, pvp_block: Dict[Union[int, str], numpy.ndarray]) -> None:
        """
        Write PVP block to the file.

        Parameters
        ----------
        pvp_block: dict
            Dictionary of `numpy.ndarray` containing the PVP arrays.
        """

        expected_channels = {c.Identifier for c in self.meta.Data.Channels}
        assert expected_channels == set(pvp_block), 'pvp_block keys do not match those in meta'
        for identifier, array in pvp_block.items():
            self.write_pvp_array(identifier, array)

    def write_signal_block(self, signal_block: Dict[Union[int, str], numpy.ndarray]) -> None:
        """
        Write signal block to the file.

        Parameters
        ----------
        signal_block: dict
            Dictionary of `numpy.ndarray` containing the signal arrays in complex64 format.
        """

        expected_channels = {c.Identifier for c in self.meta.Data.Channels}
        assert expected_channels == set(signal_block), 'signal_block keys do not match those in meta'
        for identifier, array in signal_block.items():
            self.write(array, index=identifier)

    def write_signal_block_raw(self, signal_block):
        """
        Write signal block to the file.

        Parameters
        ----------
        signal_block: dict
            Dictionary of `numpy.ndarray` containing the raw formatted
            (i.e. file storage format) signal arrays.
        """

        expected_channels = {c.Identifier for c in self.meta.Data.Channels}
        assert expected_channels == set(signal_block), 'signal_block keys do not match those in meta'
        for identifier, array in signal_block.items():
            self.write_raw(array, index=identifier)

    def write_file(
            self,
            pvp_block: Dict[Union[int, str], numpy.ndarray],
            signal_block: Dict[Union[int, str], numpy.ndarray],
            support_block: Optional[Dict[Union[int, str], numpy.ndarray]] = None):
        """
        Write the blocks to the file.

        Parameters
        ----------
        pvp_block: Dict[str, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the PVP arrays.
            Keys must be consistent with `self.meta`
        signal_block: Dict[str, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the complex64 formatted signal
            arrays.
            Keys must be consistent with `self.meta`
        support_block: None|Dict[str, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the support arrays.
        """

        self.write_pvp_block(pvp_block)
        if support_block:
            self.write_support_block(support_block)
        self.write_signal_block(signal_block)

    def write_file_raw(
            self,
            pvp_block: Dict[Union[int, str], numpy.ndarray],
            signal_block: Dict[Union[int, str], numpy.ndarray],
            support_block: Optional[Dict[Union[int, str], numpy.ndarray]] = None):
        """
        Write the blocks to the file.

        Parameters
        ----------
        pvp_block: Dict[str, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the PVP arrays.
            Keys must be consistent with `self.meta`
        signal_block: Dict[str, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the raw formatted
            (i.e. file storage format) signal arrays.
            Keys must be consistent with `self.meta`
        support_block: None|Dict[str, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the support arrays.
        """

        self.write_pvp_block(pvp_block)
        if support_block:
            self.write_support_block(support_block)
        self.write_signal_block_raw(signal_block)

    def write_chip(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Tuple[slice, ...]] = None,
            index: Union[int, str] = 0) -> None:
        self.__call__(data, start_indices=start_indices, subscript=subscript, index=index, raw=False)

    def write(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Tuple[slice, ...]] = None,
            index: Union[int, str] = 0) -> None:
        self.__call__(data, start_indices=start_indices, subscript=subscript, index=index, raw=False)

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Tuple[slice, ...]] = None,
            index: Union[int, str] = 0) -> None:
        self.__call__(data, start_indices=start_indices, subscript=subscript, index=index, raw=True)

    def __call__(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Tuple[slice, ...]] = None,
            index: Union[int, str] = 0,
            raw: bool = False) -> None:
        int_index = self._validate_channel_index(index)

        identifier = self._validate_channel_key(index)
        if not raw and not self._can_write_regular_data[identifier]:
            raise ValueError(
                'The channel `{}` has an AmpSF which has not been determined,\n\t'
                'but the corresponding PVP block has not yet been written'.format(identifier))

        BaseWriter.__call__(self, data, start_indices=start_indices, subscript=subscript, index=int_index, raw=raw)

        # check if it's fully written
        # NB: this could be refactored out, but leaving it makes the most logical
        #   sense given the pvp/support approach
        fully_written = self.data_segment[int_index].check_fully_written(warn=False)
        if fully_written:
            self.writing_details.signal_details[int_index].item_written = True

    def flush(self, force: bool = False) -> None:
        self._validate_closed()

        BaseWriter.flush(self, force=force)

        try:
            if self._in_memory:
                if self.data_segment is not None:
                    for index, entry in enumerate(self.data_segment):
                        details = self.writing_details.signal_details[index]
                        if details.item_written:
                            continue
                        if details.item_bytes is not None:
                            continue
                        if force or entry.check_fully_written(warn=force):
                            details.item_bytes = entry.get_raw_bytes(warn=False)

            self.writing_details.write_all_populated_items(self._file_object)
        except AttributeError:
            return

    def close(self):
        """
        This should perform any necessary final steps, like closing
        open file handles, deleting any temp files, etc.
        Trying to read newly created file without closing may raise a ValueError.
        """

        if hasattr(self, '_closed') and self._closed:
            return

        BaseWriter.close(self)  # NB: flush called here
        try:
            if self.writing_details is not None:
                self.writing_details.verify_all_written()
        except AttributeError:
            pass
        self._writing_details = None
        self._file_object.close()




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class CSCCGAType(TREElement):
    def __init__(self, value):
        super(CSCCGAType, self).__init__()
        self.add_field('CCG_SOURCE', 's', 18, value)
        self.add_field('REG_SENSOR', 's', 6, value)
        self.add_field('ORIGIN_LINE', 's', 7, value)
        self.add_field('ORIGIN_SAMPLE', 's', 5, value)
        self.add_field('AS_CELL_SIZE', 's', 7, value)
        self.add_field('CS_CELL_SIZE', 's', 5, value)
        self.add_field('CCG_MAX_LINE', 's', 7, value)
        self.add_field('CCG_MAX_SAMPLE', 's', 5, value)


class CSCCGA(TREExtension):
    _tag_value = 'CSCCGA'
    _data_type = CSCCGAType



"""
The detailed and involved validity checks for the sicd structure.

Note: These checks were originally implemented in the SICD component objects,
but separating this implementation is probably less confusing in the long run.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import numpy
from scipy.constants import speed_of_light
from sarpy.geometry import geocoords
from sarpy.geometry.geometry_elements import LinearRing


##############
# RgAzComp image formation parameter checks

def _rgazcomp_check_kaz_poly(
        RgAzComp,
        Timeline,
        Grid,
        SCPCOA,
        look,
        ARP_Vel) -> bool:
    """
    Check the KAZ polynomial value.

    Parameters
    ----------
    RgAzComp : sarpy.io.complex.sicd_elements.RgAzComp.RgAzCompType
    Timeline : sarpy.io.complex.sicd_elements.Timeline.TimelineType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType
    SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType
    look : int
    ARP_Vel : numpy.ndarray

    Returns
    -------
    bool
    """

    cond = True
    if Timeline.IPP is not None and len(Timeline.IPP) == 1:
        try:
            st_rate_coa = Timeline.IPP[0].IPPPoly.derivative_eval(SCPCOA.SCPTime, der_order=1)
            if Grid.Row.DeltaKCOAPoly is not None:
                krg_coa = Grid.Row.KCtr + Grid.Row.DeltaKCOAPoly.get_array(dtype='float64')
            else:
                krg_coa = Grid.Row.KCtr
            delta_kaz_per_deltav = look * krg_coa * numpy.linalg.norm(ARP_Vel) * numpy.sin(
                numpy.deg2rad(SCPCOA.DopplerConeAng)) / (SCPCOA.SlantRange * st_rate_coa)
            if isinstance(delta_kaz_per_deltav, numpy.ndarray):
                derived_kaz_poly = delta_kaz_per_deltav.dot(Timeline.IPP[0].IPPPoly.get_array(dtype='float64'))
            else:
                derived_kaz_poly = delta_kaz_per_deltav * Timeline.IPP[0].IPPPoly.get_array(dtype='float64')

            kaz_populated = RgAzComp.KazPoly.get_array(dtype='float64')
            if numpy.linalg.norm(kaz_populated - derived_kaz_poly) > 1e-3:
                RgAzComp.log_validity_error(
                    'Timeline.IPP has one element, the RgAzComp.KazPoly populated as\n\t{}\n\t'
                    'but the expected value is\n\t{}'.format(kaz_populated, derived_kaz_poly))
                cond = False
        except (AttributeError, ValueError, TypeError):
            pass
    return cond


def _rgazcomp_check_row_deltakcoa(
        RgAzComp,
        Grid,
        RadarCollection,
        ImageFormation) -> bool:
    """

    Parameters
    ----------
    RgAzComp : sarpy.io.complex.sicd_elements.RgAzComp.RgAzCompType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType
    RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
    ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

    Returns
    -------
    bool
    """

    if Grid.Row.DeltaKCOAPoly is None:
        return True

    cond = True
    row_deltakcoa = Grid.Row.DeltaKCOAPoly.get_array(dtype='float64')
    if numpy.any(row_deltakcoa != row_deltakcoa[0, 0]):
        RgAzComp.log_validity_error(
            'Grid.Row.DeltaKCOAPoly is defined, '
            'but all entries are not constant\n\t{}'.format(row_deltakcoa))
        cond = False

    if RadarCollection.RefFreqIndex is None:
        try:
            fc_proc = ImageFormation.TxFrequencyProc.center_frequency
            k_f_c = fc_proc*2/speed_of_light
            if row_deltakcoa.shape == (1, 1):
                if abs(Grid.Row.KCtr - (k_f_c - row_deltakcoa[0, 0])) > 1e-6:
                    RgAzComp.log_validity_error(
                        'the Grid.Row.DeltaCOAPoly is scalar, '
                        'and not in agreement with Grid.Row.KCtr and center frequency')
                    cond = False
            else:
                if abs(Grid.Row.KCtr - k_f_c) > 1e-6:
                    RgAzComp.log_validity_error(
                        'the Grid.Row.DeltaCOAPoly is not scalar, '
                        'and Grid.Row.KCtr not in agreement with center frequency')
                    cond = False
        except (AttributeError, ValueError, TypeError):
            pass
    return cond


def _rgazcomp_check_col_deltacoa(
        RgAzComp,
        Grid) -> bool:
    """

    Parameters
    ----------
    RgAzComp : sarpy.io.complex.sicd_elements.RgAzComp.RgAzCompType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType

    Returns
    -------
    bool
    """

    cond = True
    if Grid.Col.DeltaKCOAPoly is None:
        if Grid.Col.KCtr != 0:
            RgAzComp.log_validity_error(
                'the Grid.Col.DeltaKCOAPoly is not defined, '
                'and Grid.Col.KCtr is non-zero.')
            cond = False
    else:
        col_deltakcoa = Grid.Col.DeltaKCOAPoly.get_array(dtype='float64')
        if numpy.any(col_deltakcoa != col_deltakcoa[0, 0]):
            RgAzComp.log_validity_error(
                'the Grid.Col.DeltaKCOAPoly is defined, '
                'but all entries are not constant\n\t{}'.format(col_deltakcoa))
            cond = False

        if col_deltakcoa.shape == (1, 1) and abs(Grid.Col.KCtr + col_deltakcoa[0, 0]) > 1e-6:
            RgAzComp.log_validity_error(
                'the Grid.Col.DeltaCOAPoly is scalar, '
                'and not in agreement with Grid.Col.KCtr')
            cond = False
    return cond


def _rgazcomp_checks(the_sicd) -> bool:
    """
    Perform the RgAzComp structure validation checks.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    RgAzComp = the_sicd.RgAzComp
    Grid = the_sicd.Grid
    GeoData = the_sicd.GeoData
    SCPCOA = the_sicd.SCPCOA
    cond = True

    if Grid.ImagePlane != 'SLANT':
        the_sicd.log_validity_error(
            'The image formation algorithm is RGAZCOMP,\n\t'
            'and Grid.ImagePlane is populated as "{}",\n\t'
            'but should be "SLANT"'.format(Grid.ImagePlane))
        cond = False
    if Grid.Type != 'RGAZIM':
        the_sicd.log_validity_error(
            'The image formation algorithm is RGAZCOMP, and Grid.Type is populated as "{}",\n\t'
            'but should be "RGAZIM"'.format(Grid.Type))
        cond = False

    try:
        SCP = GeoData.SCP.ECF.get_array(dtype='float64')
        row_uvect = Grid.Row.UVectECF.get_array(dtype='float64')
        col_uvect = Grid.Col.UVectECF.get_array(dtype='float64')
        ARP_Pos = SCPCOA.ARPPos.get_array(dtype='float64')
        ARP_Vel = SCPCOA.ARPVel.get_array(dtype='float64')
        uRG = SCP - ARP_Pos
        uRG /= numpy.linalg.norm(uRG)
        left = numpy.cross(ARP_Pos / numpy.linalg.norm(ARP_Pos), ARP_Vel / numpy.linalg.norm(ARP_Vel))
        look = numpy.sign(numpy.dot(left, uRG))
        Spn = -look * numpy.cross(uRG, ARP_Vel)
        uSpn = Spn / numpy.linalg.norm(Spn)
        derived_col_vec = numpy.cross(uSpn, uRG)
    except (AttributeError, ValueError, TypeError):
        return cond

    derived_AzSF = -look * numpy.sin(numpy.deg2rad(SCPCOA.DopplerConeAng)) / SCPCOA.SlantRange
    if abs(RgAzComp.AzSF - derived_AzSF) > 1e-6:
        RgAzComp.log_validity_error(
            'AzSF is populated as {}, '
            'but expected value is {}'.format(RgAzComp.AzSF, derived_AzSF))
        cond = False

    if numpy.linalg.norm(uRG - row_uvect) > 1e-3:
        RgAzComp.log_validity_error(
            'Grid.Row.UVectECF is populated as \n\t{}\n\t'
            'which should agree with the unit range vector\n\t{}'.format(row_uvect, uRG))
        cond = False
    if numpy.linalg.norm(derived_col_vec - col_uvect) > 1e-3:
        RgAzComp.log_validity_error(
            'Grid.Col.UVectECF is populated as \n\t{}\n\t'
            'which should agree with derived vector\n\t{}'.format(col_uvect, derived_col_vec))
        cond = False

    cond &= _rgazcomp_check_kaz_poly(RgAzComp, the_sicd.Timeline, Grid, SCPCOA, look, ARP_Vel)
    cond &= _rgazcomp_check_row_deltakcoa(RgAzComp, Grid, the_sicd.RadarCollection, the_sicd.ImageFormation)
    cond &= _rgazcomp_check_col_deltacoa(RgAzComp, Grid)
    return cond


##############
# PFA image formation parameter checks

def _pfa_check_kaz_krg(PFA, Grid) -> bool:
    """
    Check the validity of the Kaz and Krg values.

    Parameters
    ----------
    PFA : sarpy.io.complex.sicd_elements.PFA.PFAType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType

    Returns
    -------
    bool
    """

    cond = True
    if PFA.STDeskew is None or not PFA.STDeskew.Applied:
        try:
            if PFA.Kaz2 - Grid.Col.KCtr > 0.5/Grid.Col.SS + 1e-10:
                PFA.log_validity_error(
                    'PFA.Kaz2 - Grid.Col.KCtr ({}) > 0.5/Grid.Col.SS ({})'.format(
                        PFA.Kaz2 - Grid.Col.KCtr, 0.5/Grid.Col.SS))
                cond = False
        except (AttributeError, TypeError, ValueError):
            pass

        try:
            if PFA.Kaz1 - Grid.Col.KCtr < -0.5/Grid.Col.SS - 1e-10:
                PFA.log_validity_error(
                    'PFA.Kaz1 - Grid.Col.KCtr ({}) < -0.5/Grid.Col.SS ({})'.format(
                        PFA.Kaz1 - Grid.Col.KCtr, -0.5/Grid.Col.SS))
                cond = False
        except (AttributeError, TypeError, ValueError):
            pass

    try:
        if PFA.Krg2 - Grid.Row.KCtr > 0.5/Grid.Row.SS + 1e-10:
            PFA.log_validity_error(
                'PFA.Krg2 - Grid.Row.KCtr ({}) > 0.5/Grid.Row.SS ({})'.format(
                    PFA.Krg2 - Grid.Row.KCtr, 0.5/Grid.Row.SS))
            cond = False
    except (AttributeError, TypeError, ValueError):
        pass

    try:
        if PFA.Krg1 - Grid.Row.KCtr < -0.5/Grid.Row.SS - 1e-10:
            PFA.log_validity_error(
                'PFA.Krg1 - Grid.Row.KCtr ({}) < -0.5/Grid.Row.SS ({})'.format(
                    PFA.Krg1 - Grid.Row.KCtr, -0.5/Grid.Row.SS))
            cond = False
    except (AttributeError, TypeError, ValueError):
        pass

    try:
        if Grid.Row.ImpRespBW > (PFA.Krg2 - PFA.Krg1) + 1e-10:
            PFA.log_validity_error(
                'Grid.Row.ImpRespBW ({}) > PFA.Krg2 - PFA.Krg1 ({})'.format(
                    Grid.Row.ImpRespBW, PFA.Krg2 - PFA.Krg1))
            cond = False
    except (AttributeError, TypeError, ValueError):
        pass

    try:
        if abs(Grid.Col.KCtr) > 1e-5 and abs(Grid.Col.KCtr - 0.5*(PFA.Kaz2 + PFA.Kaz1)) > 1e-5:
            PFA.log_validity_error(
                'Grid.Col.KCtr ({}) not within 1e-5 of 0.5*(PFA.Kaz2 + PFA.Kaz1) ({})'.format(
                    Grid.Col.KCtr, 0.5*(PFA.Kaz2 + PFA.Kaz1)))
            cond = False
    except (AttributeError, TypeError, ValueError):
        pass

    return cond


def _pfa_check_polys(PFA, Position, Timeline, SCP) -> bool:
    """

    Parameters
    ----------
    PFA : sarpy.io.complex.sicd_elements.PFA.PFAType
    Position : sarpy.io.complex.sicd_elements.Position.PositionType
    Timeline : sarpy.io.complex.sicd_elements.Timeline.TimelineType
    SCP : numpy.ndarray

    Returns
    -------
    bool
    """

    cond = True
    num_samples = max(PFA.PolarAngPoly.Coefs.size, 40)
    times = numpy.linspace(0, Timeline.CollectDuration, num_samples)

    k_a, k_sf = PFA.pfa_polar_coords(Position, SCP, times)
    if k_a is None:
        return True
    # check for agreement with k_a and k_sf derived from the polynomials
    k_a_derived = PFA.PolarAngPoly(times)
    k_sf_derived = PFA.SpatialFreqSFPoly(k_a)
    k_a_diff = numpy.amax(numpy.abs(k_a_derived - k_a))
    k_sf_diff = numpy.amax(numpy.abs(k_sf_derived - k_sf))
    if k_a_diff > 5e-3:
        PFA.log_validity_error(
            'the PolarAngPoly evaluated values do not agree with actual calculated values')
        cond = False
    if k_sf_diff > 5e-3:
        PFA.log_validity_error(
            'the SpatialFreqSFPoly evaluated values do not agree with actual calculated values')
        cond = False
    return cond


def _pfa_check_uvects(PFA, Position, Grid, SCP) -> bool:
    """

    Parameters
    ----------
    PFA : sarpy.io.complex.sicd_elements.PFA.PFAType
    Position : sarpy.io.complex.sicd_elements.Position.PositionType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType
    SCP : numpy.ndarray

    Returns
    -------
    bool
    """

    if PFA.IPN is None or PFA.FPN is None:
        return True

    cond = True
    ipn = PFA.IPN.get_array(dtype='float64')
    fpn = PFA.FPN.get_array(dtype='float64')
    row_uvect = Grid.Row.UVectECF.get_array(dtype='float64')
    col_uvect = Grid.Col.UVectECF.get_array(dtype='float64')

    pol_ref_point = Position.ARPPoly(PFA.PolarAngRefTime)
    offset = (SCP - pol_ref_point).dot(ipn)/(fpn.dot(ipn))
    ref_position_ipn = pol_ref_point + offset*fpn
    slant_range = ref_position_ipn - SCP
    u_slant_range = slant_range/numpy.linalg.norm(slant_range)
    derived_row_vector = -u_slant_range
    if numpy.linalg.norm(derived_row_vector - row_uvect) > 1e-3:
        PFA.log_validity_error(
            'the Grid.Row.UVectECF ({}) is not in good agreement with\n\t'
            'the expected value derived from PFA parameters ({})'.format(row_uvect, derived_row_vector))
        cond = False

    derived_col_vector = numpy.cross(ipn, derived_row_vector)
    if numpy.linalg.norm(derived_col_vector - col_uvect) > 1e-3:
        PFA.log_validity_error(
            'the Grid.Col.UVectECF ({}) is not in good agreement with\n\t'
            'the expected value derived from the PFA parameters ({})'.format(col_uvect, derived_col_vector))
        cond = False

    return cond


def _pfa_check_stdeskew(PFA, Grid) -> bool:
    """

    Parameters
    ----------
    PFA : sarpy.io.complex.sicd_elements.PFA.PFAType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType

    Returns
    -------
    bool
    """

    if PFA.STDeskew is None or not PFA.STDeskew.Applied:
        return True
    cond = True
    if Grid.TimeCOAPoly is not None:
        timecoa_poly = Grid.TimeCOAPoly.get_array(dtype='float64')
        if timecoa_poly.shape == (1, 1) or numpy.all(timecoa_poly.flatten()[1:] < 1e-6):
            PFA.log_validity_error(
                'PFA.STDeskew.Applied is True, and the Grid.TimeCOAPoly is essentially constant.')
            cond = False

    # the Row DeltaKCOAPoly and STDSPhasePoly should be essentially identical
    if Grid.Row is not None and Grid.Row.DeltaKCOAPoly is not None and \
            PFA.STDeskew.STDSPhasePoly is not None:
        stds_phase_poly = PFA.STDeskew.STDSPhasePoly.get_array(dtype='float64')
        delta_kcoa = Grid.Row.DeltaKCOAPoly.get_array(dtype='float64')
        rows = max(stds_phase_poly.shape[0], delta_kcoa.shape[0])
        cols = max(stds_phase_poly.shape[1], delta_kcoa.shape[1])
        exp_stds_phase_poly = numpy.zeros((rows, cols), dtype='float64')
        exp_delta_kcoa = numpy.zeros((rows, cols), dtype='float64')
        exp_stds_phase_poly[:stds_phase_poly.shape[0], :stds_phase_poly.shape[1]] = stds_phase_poly
        exp_delta_kcoa[:delta_kcoa.shape[0], :delta_kcoa.shape[1]] = delta_kcoa

        if numpy.max(numpy.abs(exp_delta_kcoa - exp_stds_phase_poly)) > 1e-6:
            PFA.log_validity_warning(
                'PFA.STDeskew.Applied is True,\n\t'
                'and the Grid.Row.DeltaKCOAPoly ({}) and PFA.STDeskew.STDSPhasePoly ({})\n\t'
                'are not in good agreement.'.format(delta_kcoa, stds_phase_poly))
            cond = False

    return cond


def _pfa_check_kctr(
        PFA,
        RadarCollection,
        ImageFormation,
        Grid) -> bool:
    """

    Parameters
    ----------
    PFA : sarpy.io.complex.sicd_elements.PFA.PFAType
    RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
    ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType

    Returns
    -------
    bool
    """

    if RadarCollection.RefFreqIndex is not None:
        return True

    cond = True
    try:
        center_freq = ImageFormation.TxFrequencyProc.center_frequency
        kap_ctr = center_freq*PFA.SpatialFreqSFPoly.Coefs[0]*2/speed_of_light
        theta = numpy.arctan(0.5*Grid.Col.ImpRespBW/Grid.Row.KCtr)  # aperture angle
        kctr_total = max(1e-2, 1 - numpy.cos(theta))  # difference between Krg and Kap
        if abs(Grid.Row.KCtr/kap_ctr - 1) > kctr_total:
            PFA.log_validity_error(
                'the Grid.Row.KCtr value ({}) is not in keeping with\n\t'
                'the expected derived from PFA parameters ({})'.format(Grid.Row.KCtr, kap_ctr))
            cond = False
    except (AttributeError, ValueError, TypeError):
        pass

    return cond


def _pfa_check_image_plane(PFA, Grid, SCPCOA, SCP) -> bool:
    """

    Parameters
    ----------
    PFA : sarpy.io.complex.sicd_elements.PFA.PFAType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType
    SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType
    SCP : numpy.ndarray

    Returns
    -------
    bool
    """

    if PFA.IPN is None or PFA.FPN is None:
        return True

    cond = True
    ipn = PFA.IPN.get_array(dtype='float64')
    fpn = PFA.FPN.get_array(dtype='float64')
    ETP = geocoords.wgs_84_norm(SCP)

    if Grid.ImagePlane == 'SLANT':
        try:
            ARP_Pos = SCPCOA.ARPPos.get_array(dtype='float64')
            ARP_Vel = SCPCOA.ARPVel.get_array(dtype='float64')
            uRG = SCP - ARP_Pos
            uRG /= numpy.linalg.norm(uRG)
            left = numpy.cross(ARP_Pos/numpy.linalg.norm(ARP_Pos), ARP_Vel/numpy.linalg.norm(ARP_Vel))
            look = numpy.sign(numpy.dot(left, uRG))
            Spn = -look*numpy.cross(uRG, ARP_Vel)
            uSpn = Spn/numpy.linalg.norm(Spn)

            if numpy.arccos(ipn.dot(uSpn)) > numpy.deg2rad(1):
                PFA.log_validity_error(
                    'the Grid.ImagePlane is "SLANT",\n\t'
                    'but COA slant plane and provided IPN are not within one degree of each other')
                cond = False
        except (AttributeError, ValueError, TypeError):
            pass
    elif Grid.ImagePlane == 'GROUND':
        if numpy.arccos(ipn.dot(ETP)) > numpy.deg2rad(3):
            PFA.log_validity_error(
                'the Grid.ImagePlane is "Ground",\n\t'
                'but the Earth Tangent Plane at SCP and provided IPN\n\t'
                'are not within three degrees of each other.')
            cond = False

    # verify that fpn points outwards
    if fpn.dot(SCP) < 0:
        PFA.log_validity_error(
            'the focus plane unit normal does not appear to be outward pointing')
        cond = False
    # check agreement between focus plane and ground plane
    if numpy.arccos(fpn.dot(ETP)) > numpy.deg2rad(3):
        PFA.log_validity_warning(
            'the focus plane unit normal is not within three degrees of the earth Tangent Plane')
    return cond


def _pfa_check_polar_angle_consistency(
        PFA,
        CollectionInfo,
        ImageFormation) -> bool:
    """

    Parameters
    ----------
    PFA : sarpy.io.complex.sicd_elements.PFA.PFAType
    CollectionInfo : sarpy.io.complex.sicd_elements.CollectionInfo.CollectionInfoType
    ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

    Returns
    -------
    bool
    """

    if CollectionInfo.RadarMode is None or CollectionInfo.RadarMode.ModeType != 'SPOTLIGHT':
        return True

    cond = True
    if PFA.Kaz1 is not None and PFA.Kaz2 is not None and PFA.Krg1 is not None:
        polar_angle_bounds = numpy.sort(PFA.PolarAngPoly(numpy.array([ImageFormation.TStartProc, ImageFormation.TEndProc], dtype='float64')))
        p2r_pol_angle_bounds = numpy.arctan(numpy.array([PFA.Kaz1, PFA.Kaz2], dtype='float64')/PFA.Krg1)
        pol_angle_bounds_diff = numpy.rad2deg(numpy.amax(numpy.abs(polar_angle_bounds - p2r_pol_angle_bounds)))
        if pol_angle_bounds_diff > 0.1:
            PFA.log_validity_warning(
                'the computed polar angle bounds ({}) from PolarAngPoly at [TStartProc, TEndProc]\n\t'
                'are not consistent with the provided polar to rectangular resampling bounds\n\t'
                '({}).'.format(polar_angle_bounds, p2r_pol_angle_bounds))
            cond = False
    return cond


def _pfa_checks(the_sicd) -> bool:
    """
    Perform the PFA structure validation checks.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    PFA = the_sicd.PFA
    Grid = the_sicd.Grid
    SCPCOA = the_sicd.SCPCOA

    cond = True
    if Grid.Type != 'RGAZIM':
        Grid.log_validity_warning(
            'The image formation algorithm is PFA,\n\t'
            'and Grid.Type is populated as "{}",\n\t'
            'but should be "RGAZIM"'.format(Grid.Type))
        cond = False
    if abs(PFA.PolarAngRefTime - SCPCOA.SCPTime) > 1e-6:
        PFA.log_validity_warning(
            'the PFA.PolarAngRefTime ({})\n\t'
            'does not agree with the SCPCOA.SCPTime ({})'.format(PFA.PolarAngRefTime, SCPCOA.SCPTime))
        cond = False
    cond &= _pfa_check_kaz_krg(PFA, Grid)
    cond &= _pfa_check_stdeskew(PFA, Grid)
    cond &= _pfa_check_kctr(PFA, the_sicd.RadarCollection, the_sicd.ImageFormation, Grid)

    try:
        SCP = the_sicd.GeoData.SCP.ECF.get_array(dtype='float64')
    except (AttributeError, ValueError, TypeError):
        return cond

    cond &= _pfa_check_polys(PFA, the_sicd.Position, the_sicd.Timeline, SCP)
    cond &= _pfa_check_uvects(PFA, the_sicd.Position, Grid, SCP)
    cond &= _pfa_check_image_plane(PFA, Grid, SCPCOA, SCP)
    cond &= _pfa_check_polar_angle_consistency(PFA, the_sicd.CollectionInfo, the_sicd.ImageFormation)
    return cond


##############
# PFA image formation parameter checks

def _rma_check_rmat(
        RMA,
        Grid,
        GeoData,
        RadarCollection,
        ImageFormation) -> bool:
    """

    Parameters
    ----------
    RMA : sarpy.io.complex.sicd_elements.RMA.RMAType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType
    GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
    RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
    ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

    Returns
    -------
    bool
    """

    cond = True
    RMAT = RMA.RMAT

    if Grid.Type != 'XCTYAT':
        Grid.log_validity_error(
            'The image formation algorithm is RMA/RMAT, which should yield '
            'Grid.Type == "XCTYAT", but Grid.Type is populated as "{}"'.format(Grid.Type))
        cond = False

    try:
        SCP = GeoData.SCP.ECF.get_array(dtype='float64')
        row_uvect = Grid.Row.UVectECF.get_array(dtype='float64')
        col_uvect = Grid.Col.UVectECF.get_array(dtype='float64')
        position_ref = RMAT.PosRef.get_array(dtype='float64')
        velocity_ref = RMAT.VelRef.get_array(dtype='float64')
        LOS = (SCP - position_ref)
        uLOS = LOS/numpy.linalg.norm(LOS)
        left = numpy.cross(
            position_ref/numpy.linalg.norm(position_ref),
            velocity_ref/numpy.linalg.norm(velocity_ref))
        look = numpy.sign(left.dot(uLOS))
        uYAT = -look*velocity_ref/numpy.linalg.norm(velocity_ref)
        uSPN = numpy.cross(uLOS, uYAT)
        uSPN /= numpy.linalg.norm(uSPN)
        uXCT = numpy.cross(uYAT, uSPN)
    except (AttributeError, ValueError, TypeError):
        return cond

    if numpy.linalg.norm(row_uvect - uXCT) > 1e-3:
        RMAT.log_validity_error(
            'the Grid.Row.UVectECF is populated as {},\n\t'
            'but derived from the RMAT parameters is expected to be {}'.format(row_uvect, uXCT))
        cond = False
    if numpy.linalg.norm(col_uvect - uYAT) > 1e-3:
        RMAT.log_validity_error(
            'the Grid.Col.UVectECF is populated as {},\n\t'
            'but derived from the RMAT parameters is expected to be {}'.format(col_uvect, uYAT))
        cond = False
    exp_doppler_cone = numpy.rad2deg(numpy.arccos(uLOS.dot(velocity_ref/numpy.linalg.norm(velocity_ref))))
    if abs(exp_doppler_cone - RMAT.DopConeAngRef) > 1e-6:
        RMAT.log_validity_error(
            'the RMAT.DopConeAngRef is populated as {},\n\t'
            'but derived from the RMAT parameters is expected to be {}'.format(RMAT.DopConeAngRef, exp_doppler_cone))
        cond = False

    if RadarCollection.RefFreqIndex is None:
        center_freq = ImageFormation.TxFrequencyProc.center_frequency
        k_f_c = center_freq*2/speed_of_light
        exp_row_kctr = k_f_c*numpy.sin(numpy.deg2rad(RMAT.DopConeAngRef))
        exp_col_kctr = k_f_c*numpy.cos(numpy.deg2rad(RMAT.DopConeAngRef))
        try:
            if abs(exp_row_kctr/Grid.Row.KCtr - 1) > 1e-3:
                RMAT.log_validity_warning(
                    'the Grid.Row.KCtr is populated as {},\n\t'
                    'and derived from the RMAT parameters is expected to be {}'.format(Grid.Row.KCtr, exp_row_kctr))
                cond = False
            if abs(exp_col_kctr/Grid.Col.KCtr - 1) > 1e-3:
                RMAT.log_validity_warning(
                    'the Grid.Col.KCtr is populated as {},\n\t'
                    'and derived from the RMAT parameters is expected to be {}'.format(Grid.Col.KCtr, exp_col_kctr))
                cond = False
        except (AttributeError, ValueError, TypeError):
            pass
    return cond


def _rma_check_rmcr(
        RMA,
        Grid,
        GeoData,
        RadarCollection,
        ImageFormation) -> bool:
    """

    Parameters
    ----------
    RMA : sarpy.io.complex.sicd_elements.RMA.RMAType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType
    GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
    RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
    ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

    Returns
    -------
    bool
    """

    cond = True
    RMCR = RMA.RMCR

    if Grid.Type != 'XRGYCR':
        Grid.log_validity_error(
            'The image formation algorithm is RMA/RMCR, which should yield '
            'Grid.Type == "XRGYCR", but Grid.Type is populated as "{}"'.format(Grid.Type))
        cond = False

    try:
        SCP = GeoData.SCP.ECF.get_array(dtype='float64')
        row_uvect = Grid.Row.UVectECF.get_array(dtype='float64')
        col_uvect = Grid.Col.UVectECF.get_array(dtype='float64')
        position_ref = RMCR.PosRef.get_array(dtype='float64')
        velocity_ref = RMCR.VelRef.get_array(dtype='float64')
        uXRG = SCP - position_ref
        uXRG /= numpy.linalg.norm(uXRG)
        left = numpy.cross(
            position_ref/numpy.linalg.norm(position_ref),
            velocity_ref/numpy.linalg.norm(velocity_ref))
        look = numpy.sign(left.dot(uXRG))
        uSPN = look*numpy.cross(velocity_ref/numpy.linalg.norm(velocity_ref), uXRG)
        uSPN /= numpy.linalg.norm(uSPN)
        uYCR = numpy.cross(uSPN, uXRG)
    except (AttributeError, ValueError, TypeError):
        return cond

    if numpy.linalg.norm(row_uvect - uXRG) > 1e-3:
        RMCR.log_validity_error(
            'the Grid.Row.UVectECF is populated as {},\n\t'
            'but derived from the RMCR parameters is expected to be {}'.format(row_uvect, uXRG))
        cond = False
    if numpy.linalg.norm(col_uvect - uYCR) > 1e-3:
        RMCR.log_validity_error(
            'the Grid.Col.UVectECF is populated as {},\n\t'
            'but derived from the RMCR parameters is expected to be {}'.format(col_uvect, uYCR))
        cond = False
    exp_doppler_cone = numpy.rad2deg(numpy.arccos(uXRG.dot(velocity_ref/numpy.linalg.norm(velocity_ref))))
    if abs(exp_doppler_cone - RMCR.DopConeAngRef) > 1e-6:
        RMCR.log_validity_error(
            'the RMCR.DopConeAngRef is populated as {},\n\t'
            'but derived from the RMCR parameters is expected to be {}'.format(RMCR.DopConeAngRef, exp_doppler_cone))
        cond = False
    if abs(Grid.Col.KCtr) > 1e-6:
        Grid.log_validity_error(
            'The image formation algorithm is RMA/RMCR,\n\t'
            'but Grid.Col.KCtr is non-zero ({}).'.format(Grid.Col.KCtr))
        cond = False

    if RadarCollection.RefFreqIndex is None:
        center_freq = ImageFormation.TxFrequencyProc.center_frequency
        k_f_c = center_freq*2/speed_of_light
        try:
            if abs(k_f_c/Grid.Row.KCtr - 1) > 1e-3:
                RMCR.log_validity_warning(
                    'the Grid.Row.KCtr is populated as {},\n\t'
                    'and derived from the RMCR parameters is expected to be {}'.format(Grid.Row.KCtr, k_f_c))
                cond = False
        except (AttributeError, ValueError, TypeError):
            pass
    return cond


def _rma_check_inca(
        RMA,
        Grid,
        GeoData,
        RadarCollection,
        CollectionInfo,
        Position) -> bool:
    """

    Parameters
    ----------
    RMA : sarpy.io.complex.sicd_elements.RMA.RMAType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType
    GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
    RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
    CollectionInfo : sarpy.io.complex.sicd_elements.CollectionInfo.CollectionInfoType
    Position :  sarpy.io.complex.sicd_elements.Position.PositionType

    Returns
    -------
    bool
    """

    cond = True
    INCA = RMA.INCA

    if Grid.Type != 'RGZERO':
        Grid.log_validity_warning(
            'The image formation algorithm is RMA/INCA, which should yield '
            'Grid.Type == "RGZERO", but Grid.Type is populated as "{}"'.format(Grid.Type))
        cond = False

    if CollectionInfo.RadarMode.ModeType == 'SPOTLIGHT':
        if INCA.DopCentroidPoly is not None:
            INCA.log_validity_error(
                'the CollectionInfo.RadarMode.ModeType == "SPOTLIGHT",\n\t'
                'and INCA.DopCentroidPoly is populated.')
            cond = False
        if INCA.DopCentroidCOA is True:
            INCA.log_validity_error(
                'the CollectionInfo.RadarMode.ModeType == "SPOTLIGHT",\n\t'
                'and INCA.DopCentroidCOA == True.')
            cond = False
    else:
        if INCA.DopCentroidPoly is None:
            INCA.log_validity_error(
                'the CollectionInfo.RadarMode.ModeType == "{}",\n\t'
                'and INCA.DopCentroidPoly is not populated.'.format(CollectionInfo.RadarMode.ModeType))
            cond = False
        if INCA.DopCentroidCOA is not True:
            INCA.log_validity_error(
                'the CollectionInfo.RadarMode.ModeType == "{}",\n\t'
                'and INCA.DopCentroidCOA is not True.'.format(CollectionInfo.RadarMode.ModeType))
            cond = False
        if Grid.Col.DeltaKCOAPoly is not None and INCA.DopCentroidPoly is not None:
            col_deltakcoa = Grid.Col.DeltaKCOAPoly.get_array(dtype='float64')
            dop_centroid = INCA.DopCentroidPoly.get_array(dtype='float64')
            rows = max(col_deltakcoa.shape[0], dop_centroid.shape[0])
            cols = max(col_deltakcoa.shape[1], dop_centroid.shape[1])
            exp_deltakcoa1 = numpy.zeros((rows, cols), dtype='float64')
            exp_deltakcoa2 = numpy.zeros((rows, cols), dtype='float64')
            exp_deltakcoa1[:col_deltakcoa.shape[0], :col_deltakcoa.shape[1]] = col_deltakcoa
            exp_deltakcoa2[:dop_centroid.shape[0], :dop_centroid.shape[1]] = dop_centroid*INCA.TimeCAPoly[1]
            if numpy.max(numpy.abs(exp_deltakcoa1 - exp_deltakcoa2)) > 1e-6:
                INCA.log_validity_error(
                    'the Grid.Col.DeltaKCOAPoly ({}),\n\t'
                    'INCA.DopCentroidPoly ({}), and INCA.TimeCAPoly ({}) '
                    'are inconsistent.'.format(col_deltakcoa,
                                               dop_centroid,
                                               INCA.TimeCAPoly.get_array(dtype='float64')))
                cond = False

    center_freq = RadarCollection.TxFrequency.center_frequency
    if abs(center_freq/INCA.FreqZero - 1) > 1e-5:
        INCA.log_validity_error(
            'the INCA.FreqZero ({}) should typically agree with center '
            'transmit frequency ({})'.format(INCA.FreqZero, center_freq))
        cond = False

    if abs(Grid.Col.KCtr) > 1e-8:
        Grid.log_validity_error(
            'The image formation algorithm is RMA/INCA, but Grid.Col.KCtr is '
            'non-zero ({})'.format(Grid.Col.KCtr))
        cond = False

    if RadarCollection.RefFreqIndex is None:
        exp_row_kctr = INCA.FreqZero*2/speed_of_light
        if abs(exp_row_kctr/Grid.Row.KCtr - 1) > 1e-8:
            INCA.log_validity_error(
                'the Grid.Row.KCtr is populated as ({}),\n\t'
                'which is not consistent with INCA.FreqZero ({})'.format(Grid.Row.KCtr, INCA.FreqZero))
            cond = False

    try:
        SCP = GeoData.SCP.ECF.get_array(dtype='float64')
        row_uvect = Grid.Row.UVectECF.get_array(dtype='float64')
        col_uvect = Grid.Col.UVectECF.get_array(dtype='float64')
        scp_time = INCA.TimeCAPoly[0]
        ca_pos = Position.ARPPoly(scp_time)
        ca_vel = Position.ARPPoly.derivative_eval(scp_time, der_order=1)
        RG = SCP - ca_pos
        uRG = RG/numpy.linalg.norm(RG)
        left = numpy.cross(ca_pos/numpy.linalg.norm(ca_pos), ca_vel/numpy.linalg.norm(ca_vel))
        look = numpy.sign(left.dot(uRG))
        uSPN = -look*numpy.cross(uRG, ca_vel)
        uSPN /= numpy.linalg.norm(uSPN)
        uAZ = numpy.cross(uSPN, uRG)
    except (AttributeError, ValueError, TypeError):
        return cond

    if numpy.linalg.norm(row_uvect - uRG) > 1e-3:
        INCA.log_validity_error(
            'the Grid.Row.UVectECF is populated as {},\n\t'
            'but derived from the INCA parameters is expected to be {}'.format(row_uvect, uRG))
        cond = False
    if numpy.linalg.norm(col_uvect - uAZ) > 1e-3:
        INCA.log_validity_error(
            'the Col.UVectECF is populated as {},\n\t'
            'but derived from the INCA parameters is expected to be {}'.format(col_uvect, uAZ))
        cond = False

    exp_R_CA_SCP = numpy.linalg.norm(RG)
    if abs(exp_R_CA_SCP - INCA.R_CA_SCP) > 1e-2:
        INCA.log_validity_error(
            'the INCA.R_CA_SCP is populated as {},\n\t'
            'but derived from the INCA parameters is expected to be {}'.format(INCA.R_CA_SCP, exp_R_CA_SCP))
        cond = False

    drate_const = INCA.DRateSFPoly[0, 0]
    exp_drate_const = 1./abs(INCA.TimeCAPoly[1]*numpy.linalg.norm(ca_vel))
    if abs(exp_drate_const - drate_const) > 1e-3:
        INCA.log_validity_error(
            'the populated INCA.DRateSFPoly constant term ({})\n\t'
            'and expected constant term ({}) are not consistent.'.format(drate_const, exp_drate_const))
        cond = False
    return cond


def _rma_checks(the_sicd) -> bool:
    """
    Perform the RMA structure validation checks.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    RMA = the_sicd.RMA

    if RMA.ImageType == 'RMAT':
        return _rma_check_rmat(RMA, the_sicd.Grid, the_sicd.GeoData, the_sicd.RadarCollection, the_sicd.ImageFormation)
    elif RMA.ImageType == 'RMCR':
        return _rma_check_rmcr(RMA, the_sicd.Grid, the_sicd.GeoData, the_sicd.RadarCollection, the_sicd.ImageFormation)
    elif RMA.ImageType == 'INCA':
        return _rma_check_inca(RMA, the_sicd.Grid, the_sicd.GeoData, the_sicd.RadarCollection, the_sicd.CollectionInfo, the_sicd.Position)
    return True


##############
# SICD checks

def _validate_scp_time(the_sicd) -> bool:
    """
    Validate the SCPTime.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if the_sicd.SCPCOA is None or the_sicd.SCPCOA.SCPTime is None or \
            the_sicd.Grid is None or the_sicd.Grid.TimeCOAPoly is None:
        return True

    cond = True
    val1 = the_sicd.SCPCOA.SCPTime
    val2 = the_sicd.Grid.TimeCOAPoly[0, 0]
    if abs(val1 - val2) > 1e-6:
        the_sicd.log_validity_error(
            'SCPTime populated as {},\n\t'
            'and constant term of TimeCOAPoly populated as {}'.format(val1, val2))
        cond = False
    return cond


def _validate_image_form_parameters(
        the_sicd,
        alg_type: str) -> bool:
    """
    Validate the image formation parameter specifics.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
    alg_type : str

    Returns
    -------
    bool
    """

    cond = True
    if the_sicd.ImageFormation.ImageFormAlgo is None:
        the_sicd.log_validity_warning(
            'Image formation algorithm(s) `{}` populated,\n\t'
            'but ImageFormation.ImageFormAlgo was not set.\n\t'
            'ImageFormation.ImageFormAlgo has been set HERE,\n\t'
            'but the incoming structure was incorrect.'.format(alg_type))
        the_sicd.ImageFormation.ImageFormAlgo = alg_type.upper()
        cond = False
    elif the_sicd.ImageFormation.ImageFormAlgo == 'OTHER':
        the_sicd.log_validity_warning(
            'Image formation algorithm `{0:s}` populated,\n\t'
            'but ImageFormation.ImageFormAlgo populated as `OTHER`.\n\t'
            'Possibly non-applicable validity checks will '
            'be performed using the `{0:s}` object'.format(alg_type))
    elif the_sicd.ImageFormation.ImageFormAlgo != alg_type:
        the_sicd.log_validity_warning(
            'Image formation algorithm {} populated,\n\t'
            'but ImageFormation.ImageFormAlgo populated as {}.\n\t'
            'ImageFormation.ImageFormAlgo has been set properly HERE,\n\t'
            'but the incoming structure was incorrect.'.format(alg_type, the_sicd.ImageFormation.ImageFormAlgo))
        the_sicd.ImageFormation.ImageFormAlgo = alg_type.upper()
        cond = False

    # verify that the referenced received channels are consistent with radar collection
    if the_sicd.ImageFormation.RcvChanProc is not None:
        channels = the_sicd.ImageFormation.RcvChanProc.ChanIndices
        if channels is None or len(channels) < 1:
            the_sicd.ImageFormation.RcvChanProc.log_validity_error('No ChanIndex values populated')
            cond = False
        else:
            rcv_channels = the_sicd.RadarCollection.RcvChannels
            if rcv_channels is None:
                the_sicd.ImageFormation.RcvChanProc.log_validity_error(
                    'Some ChanIndex values are populated,\n\t'
                    'but no RadarCollection.RcvChannels is populated.')
                cond = False
            else:
                for i, entry in enumerate(channels):
                    if not (0 < entry <= len(rcv_channels)):
                        the_sicd.ImageFormation.RcvChanProc.log_validity_error(
                            'ChanIndex entry {} is populated as {},\n\tbut must be in '
                            'the range [1, {}]'.format(i, entry, len(rcv_channels)))
                        cond = False
    if the_sicd.Grid is None:
        return cond

    if alg_type == 'RgAzComp':
        cond &= _rgazcomp_checks(the_sicd)
    elif alg_type == 'PFA':
        cond &= _pfa_checks(the_sicd)
    elif alg_type == 'RMA':
        cond &= _rma_checks(the_sicd)
    elif the_sicd.ImageFormation.ImageFormAlgo == 'OTHER':
        the_sicd.log_validity_warning(
            'Image formation algorithm populated as "OTHER", which inherently limits SICD analysis capability')
        cond = False
    return cond


def _validate_image_formation(the_sicd) -> bool:
    """
    Validate the image formation.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if the_sicd.ImageFormation is None:
        the_sicd.log_validity_error(
            'ImageFormation attribute is not populated, and ImageFormType is {}. This '
            'cannot be valid.'.format(the_sicd.ImageFormType))
        return False  # nothing more to be done.

    alg_types = []
    for alg in ['RgAzComp', 'PFA', 'RMA']:
        if getattr(the_sicd, alg) is not None:
            alg_types.append(alg)

    if len(alg_types) > 1:
        the_sicd.log_validity_error(
            'ImageFormation.ImageFormAlgo is set as {}, and multiple SICD image formation parameters {} are set.\n\t'
            'Only one image formation algorithm should be set, and ImageFormation.ImageFormAlgo '
            'should match.'.format(the_sicd.ImageFormation.ImageFormAlgo, alg_types))
        return False
    elif len(alg_types) == 0:
        if the_sicd.ImageFormation.ImageFormAlgo is None:
            the_sicd.log_validity_warning(
                'ImageFormation.ImageFormAlgo is not set, and there is no corresponding\n\t'
                'RgAzComp, PFA, or RMA SICD parameters set. Setting ImageFormAlgo to "OTHER".')
            the_sicd.ImageFormation.ImageFormAlgo = 'OTHER'
        elif the_sicd.ImageFormation.ImageFormAlgo == 'OTHER':
            the_sicd.log_validity_warning(
                'Image formation algorithm populated as "OTHER", which inherently limits SICD analysis capability')
        else:
            the_sicd.log_validity_error(
                'No RgAzComp, PFA, or RMA SICD parameters populated, but ImageFormation.ImageFormAlgo '
                'is set as {}.'.format(the_sicd.ImageFormation.ImageFormAlgo))
            return False
        return True
    # there is exactly one algorithm type populated
    return _validate_image_form_parameters(the_sicd, alg_types[0])


def _validate_antenna(the_sicd) -> bool:
    """
    Validate the Antenna Node

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """
    if the_sicd.Antenna is None:
        return True
    valid = True
    for branch_name in ['Tx', 'Rcv', 'TwoWay']:
        branch = getattr(the_sicd.Antenna, branch_name)
        if branch is None:
            continue
        for pattern_name in ['Array', 'Elem']:
            pattern = getattr(branch, pattern_name)
            if pattern is not None:
                if pattern.GainPoly is not None:
                    gbs = pattern.GainPoly(0, 0)
                    if gbs != 0.0:
                        branch.log_validity_error(f'The constant coefficient of the {branch_name}.{pattern_name}.GainPoly is {gbs} and should be 0.0')
                        valid = False
                if pattern.PhasePoly is not None:
                    pbs = pattern.PhasePoly(0, 0)
                    if pbs != 0.0:
                        branch.log_validity_error(f'The constant coefficient of the {branch_name}.{pattern_name}.PhasePoly is {pbs} and should be 0.0')
                        valid = False
        if branch.GainBSPoly is not None:
            gz = branch.GainBSPoly(0)
            if gz != 0.0:
                branch.log_validity_error(f'The constant coefficient of the {branch_name}.GainBSPoly is {gz} and should be 0.0')
    return valid


def _validate_acp(the_sicd) -> bool:
    """
    Validate the RadarCollection/Area/Corner/ACP nodes

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """
    if (the_sicd.RadarCollection is None
            or the_sicd.RadarCollection.Area is None
            or the_sicd.RadarCollection.Area.Corner is None):
        return False
    acp = the_sicd.RadarCollection.Area.Corner.get_array(dtype='float64')
    linear_ring = LinearRing(coordinates=acp)
    area = linear_ring.get_area()
    if area == 0:
        the_sicd.GeoData.log_validity_error(
            'Corner encloses no area.\n\t'
            '**disregard if crosses the +/-180 boundary')
        return False
    elif area < 0:
        the_sicd.GeoData.log_validity_error(
            "Corner must be traversed in clockwise direction.\n\t"
            "**disregard if crosses the +/-180 boundary")
        return False
    return True


def _validate_image_segment_id(the_sicd) -> bool:
    """
    Validate the image segment id.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if the_sicd.ImageFormation is None or the_sicd.RadarCollection is None:
        return False

    # get the segment identifier
    seg_id = the_sicd.ImageFormation.SegmentIdentifier
    # get the segment list
    try:
        seg_list = the_sicd.RadarCollection.Area.Plane.SegmentList
    except AttributeError:
        seg_list = None

    if seg_id is None:
        if seg_list is not None:
            the_sicd.log_validity_error(
                'ImageFormation.SegmentIdentifier is not populated, but\n\t'
                'RadarCollection.Area.Plane.SegmentList is populated.\n\t'
                'ImageFormation.SegmentIdentifier should be set to identify the appropriate segment.')
            return False
        return True

    if seg_list is None:
        the_sicd.log_validity_error(
            'ImageFormation.SegmentIdentifier is populated as {},\n\t'
            'but RadarCollection.Area.Plane.SegmentList is not populated.'.format(seg_id))
        return False

    # let's double-check that seg_id is sensibly populated
    the_ids = [entry.Identifier for entry in seg_list]
    if seg_id not in the_ids:
        the_sicd.log_validity_error(
            'ImageFormation.SegmentIdentifier is populated as {},\n\t'
            'but this is not one of the possible identifiers in the\n\t'
            'RadarCollection.Area.Plane.SegmentList definition {}.\n\t'
            'ImageFormation.SegmentIdentifier should be set to identify the '
            'appropriate segment.'.format(seg_id, the_ids))
        return False
    return True


def _validate_spotlight_mode(the_sicd) -> bool:
    """
    Validate the spotlight mode situation.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if the_sicd.CollectionInfo is None or the_sicd.CollectionInfo.RadarMode is None \
            or the_sicd.CollectionInfo.RadarMode.ModeType is None:
        return True

    if the_sicd.Grid is None or the_sicd.Grid.TimeCOAPoly is None:
        return True

    if the_sicd.CollectionInfo.RadarMode.ModeType == 'SPOTLIGHT' and \
            the_sicd.Grid.TimeCOAPoly.Coefs.shape != (1, 1):
        the_sicd.log_validity_error(
            'CollectionInfo.RadarMode.ModeType is SPOTLIGHT,\n\t'
            'but the Grid.TimeCOAPoly is not scalar - {}.\n\t'
            'This cannot be valid.'.format(the_sicd.Grid.TimeCOAPoly.Coefs))
        return False
    elif the_sicd.Grid.TimeCOAPoly.Coefs.shape == (1, 1) and \
            the_sicd.CollectionInfo.RadarMode.ModeType != 'SPOTLIGHT':
        the_sicd.log_validity_warning(
            'The Grid.TimeCOAPoly is scalar,\n\t'
            'but the CollectionInfo.RadarMode.ModeType is not SPOTLIGHT - {}.\n\t'
            'This is likely not valid.'.format(the_sicd.CollectionInfo.RadarMode.ModeType))
        return True
    return True


def _validate_valid_data(the_sicd) -> bool:
    """
    Check that either both ValidData fields are populated, or neither.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if the_sicd.ImageData is None or the_sicd.GeoData is None:
        return True

    if the_sicd.ImageData.ValidData is not None and the_sicd.GeoData.ValidData is None:
        the_sicd.log_validity_error('ValidData is populated in ImageData, but not GeoData')
        return False

    if the_sicd.GeoData.ValidData is not None and the_sicd.ImageData.ValidData is None:
        the_sicd.log_validity_error('ValidData is populated in GeoData, but not ImageData')
        return False

    if the_sicd.GeoData.ValidData is not None and the_sicd.ImageData.ValidData is not None:
        num_geo_vert = the_sicd.GeoData.ValidData.get_array().shape[0]
        num_image_vert = the_sicd.ImageData.ValidData.get_array().shape[-1]
        if num_geo_vert != num_image_vert:
            the_sicd.log_validity_error('ValidData has different number of vertices in '
                                        f'the ImageData ({num_image_vert}) '
                                        f'and GeoData ({num_geo_vert})')
    return True


def _validate_polygons(the_sicd) -> bool:
    """
    Checks that the polygons appear to be appropriate.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    def orientation(linear_ring, the_name):
        area = linear_ring.get_area()
        if area == 0:
            the_sicd.GeoData.log_validity_error(
                '{} encloses no area.\n\t'
                '**disregard if crosses the +/-180 boundary'.format(the_name))
            return False
        elif area < 0:
            the_sicd.GeoData.log_validity_error(
                "{} must be traversed in clockwise direction.\n\t"
                "**disregard if crosses the +/-180 boundary".format(the_name))
            return False
        return True

    if the_sicd.GeoData is None:
        return True

    if the_sicd.GeoData.ImageCorners is None:
        return True  # checked elsewhere

    image_corners = the_sicd.GeoData.ImageCorners.get_array(dtype='float64')

    if numpy.any(~numpy.isfinite(image_corners)):
        the_sicd.GeoData.log_validity_error('ImageCorners populated with some infinite or NaN values')
        return False

    value = True
    lin_ring = LinearRing(coordinates=image_corners)
    value &= orientation(lin_ring, 'ImageCorners')

    if the_sicd.GeoData.ValidData is not None:
        valid_data = the_sicd.GeoData.ValidData.get_array(dtype='float64')
        if numpy.all(valid_data[0] == valid_data[-1]):
            the_sicd.GeoData.log_validity_warning(
                'ValidData has the same value for the first and last point')
            return False
        if numpy.any(~numpy.isfinite(valid_data)):
            the_sicd.GeoData.log_validity_error(
                'ValidData populated with some infinite or NaN values')
            value = False
        else:
            value &= orientation(LinearRing(coordinates=valid_data), 'ValidData')
            for i, entry in enumerate(valid_data):
                contained = lin_ring.contain_coordinates(entry[0], entry[1])
                close = (lin_ring.get_minimum_distance(entry[:2]) < 1e-7)
                if not (contained or close):
                    the_sicd.GeoData.log_validity_error(
                        'ValidData entry {} is not contained ImageCorners.\n\t'
                        '**disregard if crosses the +/-180 boundary'.format(i))
                    value = False

    if not the_sicd.can_project_coordinates():
        the_sicd.log_validity_warning(
            'This sicd does not permit coordinate projection,\n\t'
            'and image corner points can not be evaluated')
        return False

    origin_loc = the_sicd.project_image_to_ground_geo([0, 0])
    if numpy.abs(origin_loc[0] - image_corners[0, 0]) > 1e-3 or numpy.abs(origin_loc[1] - image_corners[0, 1]) > 1e-3:
        the_sicd.GeoData.log_validity_error(
            'The pixel coordinate [0, 0] projects to {},\n\t'
            'which is not in good agreement with the first corner point {}'.format(origin_loc, image_corners[0]))
        value = False
    return value


def _validate_polarization(the_sicd) -> bool:
    """
    Validate the polarization.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if the_sicd.ImageFormation is None or the_sicd.ImageFormation.TxRcvPolarizationProc is None:
        return True

    if the_sicd.RadarCollection is None or the_sicd.RadarCollection.RcvChannels is None:
        return True

    pol_form = the_sicd.ImageFormation.TxRcvPolarizationProc
    rcv_pols = [entry.TxRcvPolarization for entry in the_sicd.RadarCollection.RcvChannels]
    if pol_form not in rcv_pols:
        the_sicd.log_validity_error(
            'ImageFormation.TxRcvPolarizationProc is populated as {},\n\t'
            'but it not one of the tx/rcv polarizations populated for '
            'the collect {}'.format(pol_form, rcv_pols))
        return False
    return True


def _check_deltak(the_sicd) -> bool:
    """
    Checks the deltak parameters.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if the_sicd.Grid is None:
        return True

    x_coords, y_coords = None, None
    try:
        valid_vertices = the_sicd.ImageData.get_valid_vertex_data()
        if valid_vertices is None:
            valid_vertices = the_sicd.ImageData.get_full_vertex_data()
        x_coords = the_sicd.Grid.Row.SS * (
                    valid_vertices[:, 0] - (the_sicd.ImageData.SCPPixel.Row - the_sicd.ImageData.FirstRow))
        y_coords = the_sicd.Grid.Col.SS * (
                    valid_vertices[:, 1] - (the_sicd.ImageData.SCPPixel.Col - the_sicd.ImageData.FirstCol))
    except (AttributeError, ValueError):
        pass
    return the_sicd.Grid.check_deltak(x_coords, y_coords)


def _check_projection(the_sicd) -> None:
    """
    Checks the projection ability.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
    """

    if not the_sicd.can_project_coordinates():
        the_sicd.log_validity_warning(
            'No projection can be performed for this SICD.\n'
            'In particular, no derived products can be produced.')


def _validate_radiometric(
        Radiometric,
        Grid,
        SCPCOA) -> bool:
    """
    Validate the radiometric parameters.

    Parameters
    ----------
    Radiometric : sarpy.io.complex.sicd_elements.Radiometric.RadiometricType
    Grid : sarpy.io.complex.sicd_elements.Grid.GridType
    SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType

    Returns
    -------
    bool
    """

    if Grid is None or Grid.Row is None or Grid.Col is None:
        return True

    cond = True
    area_sp = Grid.get_slant_plane_area()

    if Radiometric.RCSSFPoly is not None:
        rcs_sf = Radiometric.RCSSFPoly.get_array(dtype='float64')

        if Radiometric.BetaZeroSFPoly is not None:
            beta_sf = Radiometric.BetaZeroSFPoly.get_array(dtype='float64')
            if abs(rcs_sf[0, 0] / (beta_sf[0, 0] * area_sp) - 1) > 5e-2:
                Radiometric.log_validity_error(
                    'The BetaZeroSFPoly and RCSSFPoly are not consistent.')
                cond = False

        if SCPCOA is not None:
            if Radiometric.SigmaZeroSFPoly is not None:
                sigma_sf = Radiometric.SigmaZeroSFPoly.get_array(dtype='float64')
                mult = area_sp / numpy.cos(numpy.deg2rad(SCPCOA.SlopeAng))
                if (rcs_sf[0, 0] / (sigma_sf[0, 0] * mult) - 1) > 5e-2:
                    Radiometric.log_validity_error('The SigmaZeroSFPoly and RCSSFPoly are not consistent.')
                    cond = False

            if Radiometric.GammaZeroSFPoly is not None:
                gamma_sf = Radiometric.GammaZeroSFPoly.get_array(dtype='float64')
                mult = area_sp / (numpy.cos(numpy.deg2rad(SCPCOA.SlopeAng)) * numpy.sin(numpy.deg2rad(SCPCOA.GrazeAng)))
                if (rcs_sf[0, 0] / (gamma_sf[0, 0] * mult) - 1) > 5e-2:
                    Radiometric.log_validity_error('The GammaZeroSFPoly and RCSSFPoly are not consistent.')
                    cond = False

    if Radiometric.BetaZeroSFPoly is not None:
        beta_sf = Radiometric.BetaZeroSFPoly.get_array(dtype='float64')

        if SCPCOA is not None:
            if Radiometric.SigmaZeroSFPoly is not None:
                sigma_sf = Radiometric.SigmaZeroSFPoly.get_array(dtype='float64')
                mult = 1. / numpy.cos(numpy.deg2rad(SCPCOA.SlopeAng))
                if (beta_sf[0, 0] / (sigma_sf[0, 0] * mult) - 1) > 5e-2:
                    Radiometric.log_validity_error('The SigmaZeroSFPoly and BetaZeroSFPoly are not consistent.')
                    cond = False

            if Radiometric.GammaZeroSFPoly is not None:
                gamma_sf = Radiometric.GammaZeroSFPoly.get_array(dtype='float64')
                mult = 1. / (numpy.cos(numpy.deg2rad(SCPCOA.SlopeAng)) * numpy.sin(numpy.deg2rad(SCPCOA.GrazeAng)))
                if (beta_sf[0, 0] / (gamma_sf[0, 0] * mult) - 1) > 5e-2:
                    Radiometric.log_validity_error('The GammaZeroSFPoly and BetaZeroSFPoly are not consistent.')
                    cond = False
    return cond


def _check_radiometric_recommended(radiometric) -> None:
    """
    Checks the recommended fields for the radiometric object.

    Parameters
    ----------
    radiometric : sarpy.io.complex.sicd_elements.Radiometric.RadiometricType
    """

    for attribute in ['RCSSFPoly', 'BetaZeroSFPoly', 'SigmaZeroSFPoly', 'GammaZeroSFPoly']:
        value = getattr(radiometric, attribute)
        if value is None:
            radiometric.log_validity_warning(
                'No {} field provided, and associated RCS measurements '
                'will not be possible'.format(attribute))
    if radiometric.NoiseLevel is None:
        radiometric.log_validity_warning('No Radiometric.NoiseLevel provided, so noise estimates will not be possible.')
    elif radiometric.NoiseLevel.NoiseLevelType != 'ABSOLUTE':
        radiometric.log_validity_warning(
            'Radiometric.NoiseLevel provided are not ABSOLUTE, so noise estimates '
            'are not easily available.')


def _check_recommended_attributes(the_sicd) -> None:
    """
    Checks recommended attributes.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
    """

    if the_sicd.Radiometric is None:
        the_sicd.log_validity_warning('No Radiometric data provided.')
    else:
        _check_radiometric_recommended(the_sicd.Radiometric)

    if the_sicd.Timeline is not None and the_sicd.Timeline.IPP is None:
        the_sicd.log_validity_warning(
            'No Timeline.IPP provided, so no PRF/PRI available '
            'for analysis of ambiguities.')

    if the_sicd.RadarCollection is not None and the_sicd.RadarCollection.Area is None:
        the_sicd.log_validity_info(
            'No RadarCollection.Area provided, and some tools prefer using\n\t'
            'a pre-determined output plane for consistent product definition.')

    if the_sicd.ImageData is not None and the_sicd.ImageData.ValidData is None:
        the_sicd.log_validity_info(
            'No ImageData.ValidData is defined. It is recommended to populate\n\t'
            'this data, if validity of pixels/areas is known.')

    if the_sicd.RadarCollection is not None and the_sicd.RadarCollection.RefFreqIndex is not None:
        the_sicd.log_validity_warning(
            'A reference frequency is being used. This may affect the results of\n\t'
            'this validation test, because a number tests could not be performed.')


def detailed_validation_checks(the_sicd) -> bool:
    """
    Assembles the suite of detailed sicd validation checks.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    out = _validate_scp_time(the_sicd)
    out &= _validate_image_formation(the_sicd)
    out &= _validate_image_segment_id(the_sicd)
    out &= _validate_spotlight_mode(the_sicd)
    out &= _validate_valid_data(the_sicd)
    out &= _validate_polygons(the_sicd)
    out &= _validate_polarization(the_sicd)
    out &= _check_deltak(the_sicd)
    out &= _validate_acp(the_sicd)
    out &= _validate_antenna(the_sicd)

    if the_sicd.SCPCOA is not None:
        out &= the_sicd.SCPCOA.check_values(the_sicd.GeoData)
    if the_sicd.Radiometric is not None:
        out &= _validate_radiometric(the_sicd.Radiometric, the_sicd.Grid, the_sicd.SCPCOA)

    _check_projection(the_sicd)
    _check_recommended_attributes(the_sicd)
    return out



"""
The Antenna definition for CPHD 0.3.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List

from sarpy.io.phase_history.cphd1_elements.base import DEFAULT_STRICT, FLOAT_FORMAT
from sarpy.io.complex.sicd_elements.blocks import XYZPolyType
from sarpy.io.complex.sicd_elements.Antenna import AntParamType as AntParamTypeBase

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import SerializableDescriptor, SerializableListDescriptor, \
    FloatDescriptor


class HPBWType(Serializable):
    """
    Half power beamwidth parameters.
    """

    _fields = ('DCX', 'DCY')
    _required = _fields
    _numeric_format = {'DCX': FLOAT_FORMAT, 'DCY': FLOAT_FORMAT}
    # descriptors
    DCX = FloatDescriptor(
        'DCX', _required, strict=DEFAULT_STRICT,
        docstring='Half power beamwidth in the X-axis direction cosine '
                  '(DCX).')  # type: float
    DCY = FloatDescriptor(
        'DCY', _required, strict=DEFAULT_STRICT,
        docstring='Half power beamwidth in the Y -axis direction cosine '
                  '(DCY).')  # type: float

    def __init__(self, DCX=None, DCY=None, **kwargs):
        """

        Parameters
        ----------
        DCX : float
        DCY : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DCX = DCX
        self.DCY = DCY
        super(HPBWType, self).__init__(**kwargs)


class AntParamType(AntParamTypeBase):
    """
    The antenna parameters container.
    """

    _fields = (
        'XAxisPoly', 'YAxisPoly', 'FreqZero', 'EB', 'HPBW', 'Array', 'Elem',
        'GainBSPoly', 'EBFreqShift', 'MLFreqDilation')
    _required = ('XAxisPoly', 'YAxisPoly', 'FreqZero', )
    _numeric_format = {'FreqZero': FLOAT_FORMAT}
    # descriptors
    HPBW = SerializableDescriptor(
        'HPBW', HPBWType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, HPBWType]

    def __init__(self, XAxisPoly=None, YAxisPoly=None, FreqZero=None, EB=None,
                 HPBW=None, Array=None, Elem=None, GainBSPoly=None, EBFreqShift=None,
                 MLFreqDilation=None, **kwargs):
        """
        Parameters
        ----------
        XAxisPoly : XYZPolyType
        YAxisPoly : XYZPolyType
        FreqZero : float
        EB : None|EBType
        HPBW : None|HPBWType
        Array : None|GainPhasePolyType
        Elem : None|GainPhasePolyType
        GainBSPoly : None|Poly1DType|numpy.ndarray|list|tuple
        EBFreqShift : None|bool
        MLFreqDilation : None|bool
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.HPBW = HPBW
        super(AntParamType, self).__init__(
            XAxisPoly=XAxisPoly, YAxisPoly=YAxisPoly, FreqZero=FreqZero, EB=EB,
            Array=Array, Elem=Elem, GainBSPoly=GainBSPoly, EBFreqShift=EBFreqShift,
            MLFreqDilation=MLFreqDilation, **kwargs)


class AntennaType(Serializable):
    """
    Antenna parameters that describe antenna orientation, mainlobe steering and
    gain patterns vs. time.
    """

    _fields = ('NumTxAnt', 'NumRcvAnt', 'NumTWAnt', 'Tx', 'Rcv', 'TwoWay')
    _required = ()
    _collections_tags = {
        'Tx': {'array': False, 'child_tag': 'Tx'},
        'Rcv': {'array': False, 'child_tag': 'Rcv'},
        'TwoWay': {'array': False, 'child_tag': 'TwoWay'}}
    # descriptors
    Tx = SerializableListDescriptor(
        'Tx', AntParamType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Transmit antenna pattern parameters.'
    )  # type: Union[None, List[AntParamType]]
    Rcv = SerializableListDescriptor(
        'Rcv', AntParamType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Receive antenna pattern parameters.'
    )  # type: Union[None, List[AntParamType]]
    TwoWay = SerializableListDescriptor(
        'TwoWay', AntParamType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Two-way antenna pattern parameters.'
    )  # type: Union[None, List[AntParamType]]

    def __init__(self, Tx=None, Rcv=None, TwoWay=None, **kwargs):
        """

        Parameters
        ----------
        Tx : None|List[AntParamType]
        Rcv : None|List[AntParamType]
        TwoWay : None|List[AntParamType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Tx = Tx
        self.Rcv = Rcv
        self.TwoWay = TwoWay
        super(AntennaType, self).__init__(**kwargs)

    @property
    def NumTxAnt(self):
        """
        int: The number of transmit elements.
        """

        if self.Tx is None:
            return 0
        return len(self.Tx)

    @property
    def NumRcvAnt(self):
        """
        int: The number of receive elements.
        """

        if self.Rcv is None:
            return 0
        return len(self.Rcv)

    @property
    def NumTWAnt(self):
        """
        int: The number of two way elements.
        """

        if self.TwoWay is None:
            return 0
        return len(self.TwoWay)



"""
Module for maintaining the TRE registry
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas Mccullough"


import logging
import pkgutil
from importlib import import_module
import inspect

import sarpy._extensions
from sarpy.compliance import bytes_to_string

logger = logging.getLogger(__name__)

###############
# module variables
_TRE_Registry = {}
_parsed_package = False
_default_tre_packages = 'sarpy.io.general.nitf_elements.tres'


def register_tre(tre_type, tre_id=None, replace=False):
    """
    Register a type in the TRE registry.

    Parameters
    ----------
    tre_type : type
        A subclass of TRE
    tre_id : None|str
        The id for the type. The class name will be used if not supplied.
    replace : bool
        Should we replace if a TRE with given id if already registered?

    Returns
    -------
    None
    """

    from sarpy.io.general.nitf_elements.tres.tre_elements import TREExtension

    if not issubclass(tre_type, TREExtension):
        raise TypeError('tre_type must be a subclass of sarpy.io.general.nitf_elements.header.TRE')

    if tre_type in [TREExtension, ]:
        return

    if tre_id is None:
        tre_id = tre_type.__name__
    if not isinstance(tre_id, str):
        raise TypeError('tre_id must be a string, got type {}'.format(type(tre_id)))

    if tre_id in _TRE_Registry:
        if replace:
            logger.warning(
                'TRE with id {} is already registered.\n\t'
                'We are replacing the definition.'.format(tre_type))
        else:
            logger.warning(
                'TRE with id {} is already registered.\n\t'
                'We are NOT replacing the definition.'.format(tre_type))
            return
    _TRE_Registry[tre_id] = tre_type


def find_tre(tre_id):
    """
    Try to find a TRE with given id in our registry. Return `None` if not found.

    Parameters
    ----------
    tre_id : str|bytes

    Returns
    -------
    sarpy.io.general.nitf_elements.base.TRE|None
    """

    if not _parsed_package:
        parse_package()
        load_plugin_tres()

    if isinstance(tre_id, bytes):
        tre_id = bytes_to_string(tre_id)
    if not isinstance(tre_id, str):
        raise TypeError('tre_id must be of type string. Got {}'.format(tre_id))
    return _TRE_Registry.get(tre_id.strip(), None)


def parse_package(packages=None):
    """
    Walk the packages contained in `packages`, find all subclasses of TRE, and register them.

    Returns
    -------
    None
    """

    def evaluate(the_module):
        for element_name, element_type in inspect.getmembers(the_module, inspect.isclass):
            if issubclass(element_type, TREExtension) and element_type != TREExtension:
                register_tre(element_type, tre_id=element_name, replace=False)

    from sarpy.io.general.nitf_elements.tres.tre_elements import TREExtension

    if packages is None:
        global _parsed_package
        if _parsed_package:
            return  # already parsed the default packages
        else:
            _parsed_package = True
            packages = _default_tre_packages

    if isinstance(packages, str):
        packages = [packages, ]

    logger.info('Finding and registering TREs contained in packages {}'.format(packages))
    # walk the packages, find all subclasses of TRE, dump them into our dictionary

    for start_package in packages:
        module = import_module(start_package)
        evaluate(module)
        for details in pkgutil.walk_packages(module.__path__, start_package + '.'):
            _, module_name, is_pkg = details
            sub_module = import_module(module_name)
            evaluate(sub_module)

    logger.info('We now have {} registered TREs'.format(len(_TRE_Registry)))


def load_plugin_tres():
    """Load TREs provided by plug-ins"""
    logger.info('Loading plug-in TREs')
    for entry in sarpy._extensions.entry_points(group='sarpy.io.general.nitf_elements.tre_extension'):
        tre_class = entry.load()
        logger.info("Loading TRE: {entry.name} from {entry.module}")
        register_tre(tre_class, tre_id=entry.name, replace=False)

    logger.info('We now have {} registered TREs'.format(len(_TRE_Registry)))
